from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any

from pydantic import BaseModel


class FulltextSearchResult(BaseModel):
    """Result of a full-text search operation."""
    id: str
    score: float
    content: str
    metadata: Dict[str, Any] = {}
    artifact_id: str
    chunk_id: Optional[str] = None


class FulltextSearchResults(BaseModel):
    """Collection of full-text search results."""
    results: List[FulltextSearchResult] = []
    total: int = 0
    took: float = 0.0


class FulltextDB(ABC):
    """Abstract base class for full-text search databases.
    
    This class defines the standard interface that all full-text search database implementations
    must follow. It provides methods for storing, retrieving, and searching text content.
    """
    
    @abstractmethod
    def has_index(self, index_name: str) -> bool:
        """Check if an index exists.
        
        Args:
            index_name (str): Name of the index
            
        Returns:
            bool: True if index exists, False otherwise
        """
        pass
        
    @abstractmethod
    def delete_index(self, index_name: str):
        """Delete an index.
        
        Args:
            index_name (str): Name of the index to delete
        """
        pass
        
    @abstractmethod
    def search(
        self, 
        index_name: str, 
        query: str, 
        filter: Optional[Dict[str, Any]] = None, 
        limit: int = 10,
        offset: int = 0
    ) -> Optional[FulltextSearchResults]:
        """Search for text content based on query.
        
        Args:
            index_name (str): Name of the index
            query (str): Search query
            filter (Optional[Dict[str, Any]]): Filter conditions
            limit (int): Maximum number of results to return
            offset (int): Number of results to skip
            
        Returns:
            Optional[FulltextSearchResults]: Search results or None if index doesn't exist
        """
        pass
        
    @abstractmethod
    def query(
        self, 
        index_name: str, 
        filter: Dict[str, Any], 
        limit: Optional[int] = None,
        offset: int = 0
    ) -> Optional[FulltextSearchResults]:
        """Query items from the index based on filter.
        
        Args:
            index_name (str): Name of the index
            filter (Dict[str, Any]): Filter conditions
            limit (Optional[int]): Maximum number of results to return
            offset (int): Number of results to skip
            
        Returns:
            Optional[FulltextSearchResults]: Query results or None if index doesn't exist
        """
        pass
        
    @abstractmethod
    def get(self, index_name: str, limit: Optional[int] = None) -> Optional[FulltextSearchResults]:
        """Get all items in the index.
        
        Args:
            index_name (str): Name of the index
            limit (Optional[int]): Maximum number of results to return
            
        Returns:
            Optional[FulltextSearchResults]: All items in the index or None if index doesn't exist
        """
        pass
        
    @abstractmethod
    def insert(self, index_name: str, documents: List[Dict[str, Any]]):
        """Insert documents into the index.
        
        Args:
            index_name (str): Name of the index
            documents (List[Dict[str, Any]]): List of documents to insert
        """
        pass
        
    @abstractmethod
    def upsert(self, index_name: str, documents: List[Dict[str, Any]]):
        """Update or insert documents in the index.
        
        Args:
            index_name (str): Name of the index
            documents (List[Dict[str, Any]]): List of documents to upsert
        """
        pass
        
    @abstractmethod
    def delete(
        self,
        index_name: str,
        ids: Optional[List[str]] = None,
        filter: Optional[Dict[str, Any]] = None,
    ):
        """Delete documents from the index.
        
        Args:
            index_name (str): Name of the index
            ids (Optional[List[str]]): List of document IDs to delete
            filter (Optional[Dict[str, Any]]): Filter conditions for documents to delete
        """
        pass
        
    @abstractmethod
    def reset(self):
        """Reset the database.
        
        This will delete all indices and document entries.
        """
        pass 


from pydantic import Field, BaseModel

from workspacex.chunk.base import ChunkConfig
from workspacex.config import CHROMA_CLIENT_AUTH_PROVIDER, \
    CHROMA_CLIENT_AUTH_CREDENTIALS, CHROMA_HTTP_HOST, CHROMA_HTTP_PORT, CHROMA_HTTP_SSL, CHROMA_HTTP_HEADERS, \
    CHROMA_TENANT, CHROMA_DATABASE, CHROMA_DATA_PATH, WORKSPACEX_RERANKER_MODEL_NAME, WORKSPACEX_RERANKER_BASE_URL, \
    WORKSPACEX_RERANKER_API_KEY, WORKSPACEX_RERANKER_PROVIDER
from workspacex.config import WORKSPACEX_CHUNK_OVERLAP, WORKSPACEX_CHUNK_SIZE, WORKSPACEX_CHUNK_TEXT_SPLITTER, \
    WORKSPACEX_EMBEDDING_API_BASE_URL, WORKSPACEX_EMBEDDING_API_KEY, WORKSPACEX_EMBEDDING_BATCH_SIZE, \
    WORKSPACEX_EMBEDDING_CONTEXT_LENGTH, WORKSPACEX_EMBEDDING_DIMENSIONS, WORKSPACEX_EMBEDDING_MODEL, \
    WORKSPACEX_EMBEDDING_PROVIDER, WORKSPACEX_EMBEDDING_TIMEOUT, WORKSPACEX_ENABLE_HYBRID_SEARCH, \
    WORKSPACEX_HYBRID_SEARCH_THRESHOLD, WORKSPACEX_HYBRID_SEARCH_TOP_K, WORKSPACEX_VECTOR_DB_PROVIDER, \
    WORKSPACEX_ENABLE_CHUNKING, WORKSPACEX_ENABLE_EMBEDDING, WORKSPACEX_FULLTEXT_DB_PROVIDER, ELASTICSEARCH_HOSTS, \
    ELASTICSEARCH_USERNAME, ELASTICSEARCH_PASSWORD, \
    ELASTICSEARCH_INDEX_PREFIX, ELASTICSEARCH_NUMBER_OF_SHARDS, ELASTICSEARCH_NUMBER_OF_REPLICAS
from workspacex.embedding.base import EmbeddingsConfig
from workspacex.fulltext.factory import FulltextDBConfig
from workspacex.reranker.base import RerankConfig
from workspacex.vector.factory import VectorDBConfig


class HybridSearchConfig(BaseModel):
    enabled: bool = Field(default=False, description="enabled flag")
    top_k: int = Field(default=10, description="Top K results")
    threshold: float = Field(default=0.8, description="Threshold for similarity search")
    
    @classmethod
    def from_config(cls, config: dict):
        """
        Create a HybridSearchConfig instance from a dictionary configuration.
        
        Args:
            config (dict): Configuration dictionary containing the following keys:
                - enabled (bool): Whether hybrid search is enabled
        """
        if not config:
            return None
        return cls(
            enabled=config.get("enabled", False),
            top_k=config.get("top_k", 10),
            threshold=config.get("threshold", 0.8)
        )

class WorkspaceConfig:
    enable_hybrid_search: bool = Field(default=False, description="Enable hybrid search")
    chunk_config: ChunkConfig = Field(default=None, description="Chunk configuration")
    embedding_config: EmbeddingsConfig = Field(default=None, description="Embedding configuration")
    vector_db_config: VectorDBConfig = Field(default=None, description="Vector database configuration")
    fulltext_db_config: FulltextDBConfig = Field(default=None, description="Full-text database configuration")
    reranker_config: RerankConfig = Field(default=None, description="Reranker configuration")

    def __init__(self, chunk_config: ChunkConfig = None
                 , embedding_config: EmbeddingsConfig = None
                 , vector_db_config: VectorDBConfig = None
                 , hybrid_search_config: HybridSearchConfig = None
                 , fulltext_db_config: FulltextDBConfig = None
                 , reranker_config: RerankConfig = None
                 ):
        if chunk_config is None:
            self.chunk_config = ChunkConfig(
                enabled=WORKSPACEX_ENABLE_CHUNKING,
                provider=WORKSPACEX_CHUNK_TEXT_SPLITTER,
                chunk_size=WORKSPACEX_CHUNK_SIZE,
                chunk_overlap=WORKSPACEX_CHUNK_OVERLAP
            )
        else:
            self.chunk_config = chunk_config

        if embedding_config is None:
            self.embedding_config = EmbeddingsConfig(
                enabled=WORKSPACEX_ENABLE_EMBEDDING,
                model_name=WORKSPACEX_EMBEDDING_MODEL,
                provider=WORKSPACEX_EMBEDDING_PROVIDER,
                api_key=WORKSPACEX_EMBEDDING_API_KEY,
                base_url=WORKSPACEX_EMBEDDING_API_BASE_URL,
            )
        else:
            self.embedding_config = embedding_config

        if vector_db_config is None:
            self.embedding_config = EmbeddingsConfig(
                enabled=WORKSPACEX_ENABLE_EMBEDDING,
                model_name=WORKSPACEX_EMBEDDING_MODEL,
                provider=WORKSPACEX_EMBEDDING_PROVIDER,
                api_key=WORKSPACEX_EMBEDDING_API_KEY,
                base_url=WORKSPACEX_EMBEDDING_API_BASE_URL,
                context_length=WORKSPACEX_EMBEDDING_CONTEXT_LENGTH,
                dimensions=WORKSPACEX_EMBEDDING_DIMENSIONS,
                batch_size=WORKSPACEX_EMBEDDING_BATCH_SIZE,
                timeout=WORKSPACEX_EMBEDDING_TIMEOUT
            )
        else:
            self.embedding_config = embedding_config

        if vector_db_config is None:
            self.vector_db_config = VectorDBConfig(
                provider=WORKSPACEX_VECTOR_DB_PROVIDER,
                config={
                    "allow_reset": True,
                    "anonymized_telemetry": False,
                    "chroma_client_auth_provider": CHROMA_CLIENT_AUTH_PROVIDER if CHROMA_CLIENT_AUTH_PROVIDER else None,
                    "chroma_client_auth_credentials": CHROMA_CLIENT_AUTH_CREDENTIALS if CHROMA_CLIENT_AUTH_CREDENTIALS else None,
                    "http_host": CHROMA_HTTP_HOST if CHROMA_HTTP_HOST else None,
                    "http_port": CHROMA_HTTP_PORT if CHROMA_HTTP_PORT else None,
                    "http_ssl": CHROMA_HTTP_SSL if CHROMA_HTTP_SSL else None,
                    "http_headers": CHROMA_HTTP_HEADERS if CHROMA_HTTP_HEADERS else None,
                    "tenant": CHROMA_TENANT if CHROMA_TENANT else None,
                    "database": CHROMA_DATABASE if CHROMA_DATABASE else None,
                    "data_path": CHROMA_DATA_PATH if CHROMA_DATA_PATH else None
                }
            )
        else:
            self.vector_db_config = vector_db_config

        if hybrid_search_config is None:
            self.hybrid_search_config = HybridSearchConfig(
                enabled=WORKSPACEX_ENABLE_HYBRID_SEARCH,
                top_k=WORKSPACEX_HYBRID_SEARCH_TOP_K,
                threshold=WORKSPACEX_HYBRID_SEARCH_THRESHOLD
            )
        else:
            self.hybrid_search_config = hybrid_search_config

        if fulltext_db_config is None:
            # Parse Elasticsearch hosts and remove http:// prefix if present
            hosts = ELASTICSEARCH_HOSTS.split(",") if "," in ELASTICSEARCH_HOSTS else ELASTICSEARCH_HOSTS
            cleaned_hosts = []
            if isinstance(hosts,list) and len(hosts) > 1:
                for host in hosts:
                    cleaned_hosts.append(host)
            else:
                cleaned_hosts = hosts
            
            self.fulltext_db_config = FulltextDBConfig(
                provider=WORKSPACEX_FULLTEXT_DB_PROVIDER,
                config={
                    "hosts": cleaned_hosts,
                    "username": ELASTICSEARCH_USERNAME if ELASTICSEARCH_USERNAME else None,
                    "password": ELASTICSEARCH_PASSWORD if ELASTICSEARCH_PASSWORD else None,
                    "index_prefix": ELASTICSEARCH_INDEX_PREFIX,
                    "number_of_shards": ELASTICSEARCH_NUMBER_OF_SHARDS,
                    "number_of_replicas": ELASTICSEARCH_NUMBER_OF_REPLICAS
                }
            )
        else:
            self.fulltext_db_config = fulltext_db_config

        if reranker_config is None:
            self.reranker_config = RerankConfig(
                provider=WORKSPACEX_RERANKER_PROVIDER,
                base_url=WORKSPACEX_RERANKER_BASE_URL,
                api_key=WORKSPACEX_RERANKER_API_KEY,
                model_name=WORKSPACEX_RERANKER_MODEL_NAME,
            )
        else:
            self.reranker_config = reranker_config

    @classmethod
    def from_config(cls, config: dict):
        return cls(
            chunk_config=ChunkConfig.from_config(config.get("chunk_config")) if config else None,
            embedding_config=EmbeddingsConfig.from_config(config.get("embedding_config")) if config else None,
            vector_db_config=VectorDBConfig.from_config(config.get("vector_db_config")) if config else None,
            hybrid_search_config=HybridSearchConfig.from_config(config.get("hybrid_search_config")) if config else None,
            fulltext_db_config=FulltextDBConfig.from_config(config.get("fulltext_db_config")) if config else None,
            reranker_config=RerankConfig.from_config(config.get("reranker_config")) if config else None
        )

    def to_dict(self):
        return {
            "chunk_config": self.chunk_config.model_dump(),
            "embedding_config": self.embedding_config.model_dump(),
            "vector_db_config": self.vector_db_config.model_dump(),
            "hybrid_search_config": self.hybrid_search_config.model_dump(),
            "fulltext_db_config": self.fulltext_db_config.model_dump(),
            "reranker_config": self.reranker_config.model_dump()
        }



from typing import List, Optional, Dict, Any

from workspacex.reranker.base import BaseRerankRunner, RerankConfig, RerankResult
from workspacex.artifact import Artifact
from workspacex.utils.logger import logger


class Qwen3RerankerRunner(BaseRerankRunner):
    """
    Local reranker using Qwen3-Reranker model via transformers.
    Requires transformers>=4.51.0 and torch.
    """
    def __init__(self, config: RerankConfig) -> None:
        """
        Initialize Qwen3RerankerRunner.
        Args:
            config (RerankConfig): Configuration for rerank model.
        Raises:
            ImportError: If required dependencies (torch, transformers) are not installed.
        """
        try:
            import torch
            from transformers import AutoTokenizer, AutoModelForCausalLM
        except ImportError as e:
            logger.error("Required dependencies not installed. Please install torch and transformers>=4.51.0")
            raise ImportError("Required dependencies not installed. Please install torch and transformers>=4.51.0") from e

        self.config = config
        self.model_id = config.model_name or "Qwen/Qwen3-Reranker-0.6B"

        # Initialize tokenizer and model
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id,
                                                      padding_side='left')
        self.model = AutoModelForCausalLM.from_pretrained(self.model_id)

        # Enable flash attention if available
        if torch.cuda.is_available():
            try:
                self.model = AutoModelForCausalLM.from_pretrained(
                    self.model_id,
                    torch_dtype=torch.float16,
                    attn_implementation="flash_attention_2").cuda()
            except Exception as e:
                logger.warning(
                    f"Failed to enable flash attention: {e}. Using default model."
                )
                self.model = self.model.cuda()
        self.model.eval()

        # Setup tokens and constants
        self.token_false_id = self.tokenizer.convert_tokens_to_ids("no")
        self.token_true_id = self.tokenizer.convert_tokens_to_ids("yes")
        self.max_length = 8192

        # Setup prompt templates
        self.prefix = "<|im_start|>system\nJudge whether the Document meets the requirements based on the Query and the Instruct provided. Note that the answer can only be \"yes\" or \"no\".<|im_end|>\n<|im_start|>user\n"
        self.suffix = "<|im_end|>\n<|im_start|>assistant\n<think>\n\n</think>\n\n"
        self.prefix_tokens = self.tokenizer.encode(self.prefix,
                                                  add_special_tokens=False)
        self.suffix_tokens = self.tokenizer.encode(self.suffix,
                                                  add_special_tokens=False)

        # Default instruction
        self.instruction = 'Given a web search query, retrieve relevant passages that answer the query'

    def format_instruction(self, query: str, doc: str) -> str:
        """
        Format the instruction, query and document into the required prompt format.
        Args:
            query (str): The search query
            doc (str): The document text
        Returns:
            str: Formatted prompt
        """
        return f"<Instruct>: {self.instruction}\n<Query>: {query}\n<Document>: {doc}"

    def process_inputs(self, pairs: List[str]) -> Dict[str, Any]:
        """
        Process and tokenize the input pairs.
        Args:
            pairs (List[str]): List of formatted instruction-query-document pairs
        Returns:
            Dict[str, Any]: Processed inputs ready for model
        """
        inputs = self.tokenizer(pairs,
                               padding=False,
                               truncation='longest_first',
                               return_attention_mask=False,
                               max_length=self.max_length -
                               len(self.prefix_tokens) -
                               len(self.suffix_tokens))
        for i, ele in enumerate(inputs['input_ids']):
            inputs['input_ids'][
                i] = self.prefix_tokens + ele + self.suffix_tokens
        inputs = self.tokenizer.pad(inputs,
                                   padding=True,
                                   return_tensors="pt",
                                   max_length=self.max_length)
        
        try:
            import torch
            if torch.cuda.is_available():
                for key in inputs:
                    inputs[key] = inputs[key].to(self.model.device)
        except ImportError:
            logger.error("torch not found when trying to move tensors to GPU")
            raise
        return inputs

    def compute_logits(self, inputs: Dict[str, Any]) -> List[float]:
        """
        Compute relevance scores using the model.
        Args:
            inputs (Dict[str, Any]): Processed inputs
        Returns:
            List[float]: List of relevance scores
        """
        try:
            import torch
            with torch.no_grad():
                batch_scores = self.model(**inputs).logits[:, -1, :]
                true_vector = batch_scores[:, self.token_true_id]
                false_vector = batch_scores[:, self.token_false_id]
                batch_scores = torch.stack([false_vector, true_vector], dim=1)
                batch_scores = torch.nn.functional.log_softmax(batch_scores, dim=1)
                scores = batch_scores[:, 1].exp().tolist()
                return scores
        except ImportError:
            logger.error("torch not found when computing logits")
            raise

    def run(
        self,
        query: str,
        documents: List[Artifact],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
        user: Optional[str] = None,
    ) -> List[RerankResult]:
        """
        Run rerank using Qwen3-Reranker model locally.
        Args:
            query (str): Search query.
            documents (List[Artifact]): Documents for reranking.
            score_threshold (Optional[float]): Score threshold.
            top_n (Optional[int]): Top n results.
            user (Optional[str]): Unique user id if needed.
        Returns:
            List[RerankResult]: List of rerank results.
        Raises:
            ImportError: If required dependencies are not installed.
        """
        # Handle empty documents list
        if not documents:
            return []
        
        # Format inputs
        pairs = [
            self.format_instruction(query, doc.get_reranked_text())
            for doc in documents
        ]

        # Process and get scores
        inputs = self.process_inputs(pairs)
        scores = self.compute_logits(inputs)

        # Create results
        rerank_results = []
        for idx, score in enumerate(scores):
            if score_threshold is not None and score < score_threshold:
                continue
            rerank_results.append(
                RerankResult(artifact=documents[idx], score=score))

        # Sort and filter results
        rerank_results = sorted(rerank_results,
                               key=lambda x: x.score,
                               reverse=True)
        if top_n is not None:
            rerank_results = rerank_results[:top_n]

        return rerank_results



import os

from workspacex import WorkSpace
from workspacex.base import WorkspaceConfig


async def load_workspace(workspace_id, workspace_type, storage_path, config: WorkspaceConfig = None) -> WorkSpace:
    if workspace_type == "local":
        parent_dir = os.environ.get("WORKSPACEX_LOCAL_BASE_DIR", "data/workspaces")
        return WorkSpace.from_local_storages(workspace_id=workspace_id, storage_path=parent_dir + "/" + workspace_id, config = config)
    elif workspace_type == "s3":
        return WorkSpace.from_s3_storages(workspace_id=workspace_id, storage_path=storage_path, config = config)
    else:
        raise ValueError(f"not support workspace#{workspace_type}: {workspace_id}")

def sync_load_workspace(workspace_id, workspace_type, storage_path, config: WorkspaceConfig = None) -> WorkSpace:
    if workspace_type == "local":
        parent_dir = os.environ.get("WORKSPACEX_LOCAL_BASE_DIR", "data/workspaces")
        return WorkSpace.from_local_storages(workspace_id=workspace_id, storage_path=parent_dir + "/" + workspace_id, config = config)
    elif workspace_type == "s3":
        return WorkSpace.from_s3_storages(workspace_id=workspace_id, storage_path=storage_path, config = config)
    else:
        raise ValueError(f"not support workspace#{workspace_type}: {workspace_id}")



import uuid
from typing import Any, Optional, Dict, List

from pydantic import Field

from workspacex.artifact import Artifact, ArtifactType

CODE_FILE_EXTENSION_MAP = {
                "python": "py",
                "java": "java",
                "javascript": "js",
                "typescript": "ts",
                "html": "html",
                "css": "css",
                "c": "c",
                "cpp": "cpp",
                "csharp": "cs",
                "go": "go",
                "rust": "rs",
                "ruby": "rb",
                "php": "php",
                "swift": "swift",
                "kotlin": "kt",
                "scala": "scala",
                "markdown": "md",
                "txt": "txt",
                "shell": "sh",
                "bash": "sh",
                "sh": "sh",
                "zsh": "zsh",
                "powershell": "ps1",
                "cmd": "cmd",
                "bat": "bat"
            }

class CodeArtifact(Artifact):
    code_interceptor: Any = Field(default=None, description="code executor type")

    def __init__(self, artifact_type: ArtifactType, content: Any, code_type: str, code_version: str,
                 code_interceptor_provider: Optional[str] = None,
                 artifact_id: Optional[str] = None, render_type: Optional[str] = None, **kwargs):
        # Extract filename from the first line of the content
        filename = self.extract_filename(content)
        
        # Initialize metadata, including any passed in kwargs
        metadata = {
            "code_type": code_type,
            "code_version": code_version,
            "code_interceptor_provider": code_interceptor_provider,
            "filename": filename  # Store filename in metadata
        }
        
        # Merge additional metadata from kwargs if provided
        if 'metadata' in kwargs:
            metadata.update(kwargs['metadata'])
            del kwargs['metadata']  # Remove metadata from kwargs to avoid multiple values

        super().__init__(artifact_type, content, metadata, artifact_id, render_type, **kwargs)
        self.code_interceptor = self.init_code_interceptor(code_interceptor_provider)

    @staticmethod
    def extract_filename(content: Any) -> Optional[str]:
        """Extract filename from the first line of the code block comment."""
        if isinstance(content, str):
            lines = content.splitlines()
            if lines:
                first_line = lines[0].strip()
                # Check if the first line is a shebang for bash or other interpreters
                if first_line in ["# /bin/bash", "#!/bin/bash", "#!/usr/bin/env bash", 
                                   "#!/bin/sh", "#!/usr/bin/env python", 
                                   "#!/usr/bin/env python3"]:
                    return None  # Do not return a filename
                # Check for common comment styles in various languages
                if first_line.startswith("#"):  # Python, Ruby, Shell
                    return first_line[1:].strip()  # Remove the comment symbol
                elif first_line.startswith("//"):  # Java, JavaScript, C, C++
                    return first_line[2:].strip()  # Remove the comment symbol
                elif first_line.startswith("/*") and "*/" in first_line:  # C, C++
                    return first_line.split("*/")[0][2:].strip()  # Remove comment symbols
                elif first_line.startswith("<!--"):  # HTML
                    return first_line[4:].strip()  # Remove the comment symbol
                # Add more languages as needed
        return None  # Return None if filename is unknown

    @classmethod
    def create_artifact(cls, artifact_type: ArtifactType,
                       content: Any,
                       code_type: str,
                       code_version: str,
                       code_interceptor_provider: Optional[str] = None,
                       artifact_id: Optional[str] = None,
                       render_type: Optional[str] = None,
                       **kwargs) -> "CodeArtifact":
        """Factory method to create appropriate artifact based on language"""
        if code_type in ['shell', 'sh', 'bash', 'zsh']:
            return ShellArtifact(
                artifact_type=artifact_type,
                content=content,
                code_version=code_version,
                code_interceptor_provider=code_interceptor_provider,
                artifact_id=artifact_id,
                render_type=render_type,
                **kwargs
            )
        return cls(
            artifact_type=artifact_type,
            content=content,
            code_type=code_type,
            code_version=code_version,
            code_interceptor_provider=code_interceptor_provider,
            artifact_id=artifact_id,
            render_type=render_type,
            **kwargs
        )

    @classmethod
    def from_code_content(cls, artifact_type: ArtifactType,
                          content: Any,
                          render_type: Optional[str] = None,
                          **kwargs) -> List["CodeArtifact"]:
        code_blocks = cls.extract_model_output_to_code_content(content)  # Extract code blocks
        artifacts = []  # List to store CodeArtifact instances

        for block in code_blocks:
            code_type = block['language']
            code_version = "1.0"

            if code_type in ['python', 'javascript', 'java']:
                code_interceptor_provider = "default_interceptor"
            elif code_type in ['shell', 'sh', 'bash', 'zsh']:
                code_interceptor_provider = "shell_interceptor"
            else:
                code_interceptor_provider = "generic_interceptor"

            artifact = cls.create_artifact(
                artifact_type=artifact_type,
                content=block['content'],
                code_type=code_type,
                code_version=code_version,
                code_interceptor_provider=code_interceptor_provider,
                artifact_id=block['artifact_id'],  # Use extracted artifact_id
                render_type=render_type,
                **kwargs
            )
            artifacts.append(artifact)  # Add to the list

        return artifacts  # Return the list of CodeArtifact instances

    def init_code_interceptor(self, code_interceptor_provider):
        pass

    @classmethod
    def extract_model_output_to_code_content(cls, content):
        """
        Extract code blocks from markdown content using mistune.
        
        First extracts all code blocks enclosed in triple backticks,
        then determines the language for each block.
        """

        try:
            import mistune
        except ImportError:
            # install mistune
            import subprocess
            subprocess.run(["pip", "install", "mistune>=3.0.0"], check=True)
            import mistune

        code_blocks = []
        
        #
        extracted_blocks = []
        
        # create custom Render
        class CustomRenderer(mistune.HTMLRenderer):
            def block_code(self, code, info=None):
                language = info.split()[0] if info else 'unknown'
                extracted_blocks.append({
                    "content": code,
                    "language": language
                })
                return ""

        # create Markdown render
        renderer = CustomRenderer()
        markdown = mistune.create_markdown(
            renderer=renderer
        )

        # resolve markdown
        markdown(content)
        
        # process codeblocks
        for block in extracted_blocks:
            artifact_id = str(uuid.uuid4())
            language = block['language']
            file_suffix = CODE_FILE_EXTENSION_MAP.get(language, "txt")
            
            code_blocks.append({
                "artifact_id": artifact_id,
                "content": block['content'],
                "language": language,
                "file_suffix": file_suffix
            })
        
        return code_blocks


class ShellArtifact(CodeArtifact):
    shell_result: str = Field(default="", description="shell execution result")

    def __init__(self, artifact_type: ArtifactType, content: Any, code_version: str,
                 code_interceptor_provider: Optional[str] = None,
                 artifact_id: Optional[str] = None, render_type: Optional[str] = None,
                 shell_result: str = "", **kwargs):

        code_type = "shell"

        # extract filename
        filename = self.extract_filename(content)
        
        # default set terminal.txt
        if not filename:
            filename = "terminal.txt"

        # update metadata
        metadata = kwargs.get('metadata', {})
        metadata['filename'] = filename

        # setting code_interceptor_provider
        if code_interceptor_provider is None:
            code_interceptor_provider = "shell_interceptor"

        super().__init__(artifact_type, content, code_type, code_version,
                         code_interceptor_provider, artifact_id, render_type, metadata=metadata, **kwargs)
        self.shell_result = shell_result

    def execute(self):
        # todo add
        pass



import os
from typing import Any, List, Optional
from workspacex.artifact import Artifact, ArtifactType
from workspacex.extractor.base import BaseExtractor


class NovalExtractor(BaseExtractor):
    
    """
    Extractor for Noval files.
    Loads a novel file, splits it by chapters, 
    """
    
    def extract(self, content: Any) -> Optional[list[Artifact]]:
        """Extract artifacts from content."""
        novel_file_path = content

        chapters, chapter_titles = self._load_and_split_novel(novel_file_path)

        artifacts = []
        # Create subartifacts for each chapter
        for idx, (title, chapter) in enumerate(zip(chapter_titles, chapters)):
            sub_meta = {
                "chapter_title": title,
                "chapter_index": idx + 1
            }
            subartifact = Artifact(
                artifact_id=f"chapter_{idx+1}_{title.replace(' ', '')}",
                artifact_type=ArtifactType.NOVEL,
                content=chapter,
                metadata=sub_meta
            )
            artifacts.append(subartifact)
        return artifacts

    async def async_extract(self, content: Any, **kwargs) -> Optional[list[Artifact]]:
        pass

    def _load_and_split_novel(self, novel_file_path: str) -> tuple[List[str], List[str]]:
        """
        Load the novel file and split it into chapters.
        Args:
            novel_file_path: Path to the novel file
        Returns:
            Tuple of (full content, list of chapters, list of chapter titles)
        """
        with open(novel_file_path, 'r', encoding='utf-8') as f:
            content = f.read()
        import re
        # Split by chapter heading, e.g., "第001章" or "第1章"
        pattern = r'(第[0-9一二三四五六七八九十百千]+章[\s\S]*?)(?=第[0-9一二三四五六七八九十百千]+章|$)'
        chapters = re.findall(pattern, content)
        # Extract chapter titles
        title_pattern = r'(第[0-9一二三四五六七八九十百千]+章[^\n]*)'
        chapter_titles = re.findall(title_pattern, content)
        return chapters, chapter_titles
    
noval_extractor = NovalExtractor()


import logging
import os
import time
from typing import Any

from langchain_core.runnables import RunnableConfig
from langchain_openai import ChatOpenAI


class LangFuseHolder:

    def __init__(self):

        if os.environ.get("LANGFUSE_ENABLED", "False").lower() == "true":
            from langfuse.langchain import CallbackHandler
            self._langfuse_handler = CallbackHandler()

    def get_handler(self):
        return self._langfuse_handler

langfuse = LangFuseHolder()

def call_llm(prompt: str, model_name: str = None, llm_config: dict[str, Any] = {}) -> str:

    if os.environ.get('LLM_API_KEY') is None:
        raise ValueError("LLM_API_KEY is not set")
    if os.environ.get('LLM_BASE_URL') is None:
        raise ValueError("LLM_BASE_URL is not set")
    
    try:
        start_time = time.time()
        llm_model = get_llm_model(model_name, llm_config)
        if os.environ.get("LANGFUSE_ENABLED", "False").lower() == "true":
            response = llm_model.invoke(
                [{"role": "user", "content": prompt}],
                config=RunnableConfig(callbacks=[langfuse.get_handler()]))
        else:
            response = llm_model.invoke([{"role": "user", "content": prompt}])
        use_time = time.time() - start_time
        logging.info(f"LLM response[{len(prompt)} chars -> use {use_time:.2f} s] result is: {response.content} 🤖")
        return response.content
    except Exception as e:    
        logging.error(f"Failed to call LLM: {e}")
        raise ValueError(f"Failed to call LLM model: {e}")
    


async def call_llm_async(prompt: str, model_name: str = None, llm_config=None, system_prompt: str = None) -> str:
    if llm_config is None:
        llm_config = {}

    if os.environ.get('LLM_API_KEY') is None:
        raise ValueError("LLM_API_KEY is not set")
    if os.environ.get('LLM_BASE_URL') is None:
        raise ValueError("LLM_BASE_URL is not set")
    
    try:
        start_time = time.time()
        llm_model = get_llm_model(model_name, llm_config)
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})

        log_msg = f"📝 LLM[{llm_model.model_name}] request prompt:\n"
        for idx, msg in enumerate(messages):
            role = msg.get("role", "unknown")
            content = msg.get("content", "")
            log_msg += f"  [{idx}] ({role}): {content}\n"
        logging.info(log_msg)
        if os.environ.get("LANGFUSE_ENABLED", "False").lower() == "true":
            response = await llm_model.ainvoke(
                messages,
                config=RunnableConfig(callbacks=[langfuse.get_handler()]))
        else:
            response = await llm_model.ainvoke(
                messages
            )
        use_time = time.time() - start_time
        logging.info(f"LLM response[{len(prompt)} chars -> use {use_time:.2f} s] result is: {response.content} 🤖 -> {response.usage_metadata}")
        return response.content
    except Exception as e:    
        logging.error(f"Failed to call LLM: {e}")
        raise ValueError(f"Failed to call LLM model: {e}")

def get_llm_model(model_name: str = None, llm_config=None) -> ChatOpenAI:
    if llm_config is None:
        llm_config = {}
    if os.environ.get('LLM_API_KEY') is None:
        raise ValueError("LLM_API_KEY is not set")
    if os.environ.get('LLM_BASE_URL') is None:
        raise ValueError("LLM_BASE_URL is not set")
    
    return ChatOpenAI(
        api_key=os.environ.get('LLM_API_KEY'),
        base_url=os.environ.get('LLM_BASE_URL'),
        model=model_name if model_name else os.environ.get('LLM_MODEL'),
        timeout=llm_config.get('timeout', 60),
        max_retries=llm_config.get('max_retries', 2),
        max_tokens=llm_config.get('max_tokens', 4096),
        temperature=llm_config.get('temperature', 0.5),
        frequency_penalty=llm_config.get('frequency_penalty', 0),
        presence_penalty=llm_config.get('presence_penalty', 0),
        extra_body=llm_config.get('extra_body', {})
    )





from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseModel, Field

from workspacex.artifact import Artifact


class RerankConfig(BaseModel):
    provider: str = Field(default="local", description="Provider")
    base_url: str = Field(..., description="Base URL")
    api_key: Optional[str] = Field(default=None, description="API key")
    model_name: Optional[str] = Field(default=None, description="Model name")
    
    @staticmethod
    def from_config(config: dict) -> "RerankConfig":
        if not config:
            return None
        return RerankConfig(**config)



class RerankResult(BaseModel):
    artifact: Artifact = Field(..., description="Artifact")
    score: float = Field(..., description="Score")


class BaseRerankRunner(ABC):
    @abstractmethod
    def run(
            self,
            query: str,
            documents: list[Artifact],
            score_threshold: Optional[float] = 0.8,
            top_n: Optional[int] = 5,
            user: Optional[str] = None,
    ) -> list[RerankResult]:
        """
        Run rerank model
        :param query: search query
        :param documents: documents for reranking
        :param score_threshold: score threshold
        :param top_n: top n
        :param user: unique user id if needed

                results = reranker.run("你好", documents=[
                    Artifact(artifact_type=ArtifactType.TEXT, content="你好"),
                    Artifact(artifact_type=ArtifactType.TEXT, content="我很好"),
                    Artifact(artifact_type=ArtifactType.TEXT, content="谢谢你"),
                    Artifact(artifact_type=ArtifactType.TEXT, content="很高兴遇到你"),
                ])
                for result in results:
                    print(f"{result.artifact.content}: {result.score}")
                # 你好: 0.989096999168396
                # 很高兴遇到你: 0.30921047925949097
                # 谢谢你: 0.10078048706054688
                # 我很好: 0.06917418539524078
        """
        raise NotImplementedError



from typing import List

from langchain_text_splitters import CharacterTextSplitter

from workspacex.artifact import Artifact
from .base import Chunk, ChunkConfig, ChunkerBase


class CharacterChunker(ChunkerBase):
    """
    A chunker that splits text by characters.
    """

    def __init__(self, config: ChunkConfig):
        super().__init__(config=config)
        self._text_splitter = CharacterTextSplitter(
            separator="\n",
            chunk_size=self.config.chunk_size,
            chunk_overlap=self.config.chunk_overlap,
            length_function=len,
            is_separator_regex=False,
        )

    async def chunk(self, artifact: Artifact) -> List[Chunk]:
        """
        Chunks the given content by characters.

        Args:
            content: The string content to chunk.

        Returns:
            A list of `Chunk` objects.
        """
        texts = self._text_splitter.split_text(artifact.content)

        return self._create_chunks(texts, artifact)


# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# reference https://github.com/google/langextract/blob/main/langextract/io.py

"""Supports Input and Output Operations for Data Annotations."""

from typing import Tuple

import requests

from workspacex.utils import progress

DEFAULT_TIMEOUT_SECONDS = 30


class InvalidDatasetError(Exception):
    """Error raised when Dataset is empty or invalid."""


def is_url(text: str) -> bool:
    """Check if the given text is a URL.

    Args:
      text: The string to check.

    Returns:
      True if the text is a URL (starts with http:// or https://), False
      otherwise.
    """
    return text.startswith('http://') or text.startswith('https://')


async def download_pdf_from_url(
        url: str,
        timeout: int = DEFAULT_TIMEOUT_SECONDS,
        show_progress: bool = True,
        chunk_size: int = 8192,
) -> Tuple[str, str]:
    """
    download pdf from url, return temp file path

    """

    try:
        # Make initial request to get headers
        response = requests.get(url, stream=True, timeout=timeout)
        response.raise_for_status()

        # Get content length for progress bar
        total_size = int(response.headers.get('Content-Length', 0))

        filename = url.split('/')[-1][:50]

        # save as pdf temp file return file path
        temp_file_path = f"/tmp/{filename}.pdf"
        # Download content with progress bar
        chunks = []
        if show_progress and total_size > 0:
            progress_bar = progress.create_download_progress_bar(
                total_size=total_size, url=url
            )


            with open(temp_file_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:
                        f.write(chunk)
                        progress_bar.update(len(chunk))
            progress_bar.close()
        else:
            # Download without progress bar
            with open(temp_file_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:
                        f.write(chunk)

        return f"{filename}.pdf",temp_file_path

    except requests.RequestException as e:
        raise requests.RequestException(
            f'Failed to download from {url}: {str(e)}'
        ) from e
    pass


def download_text_from_url(
        url: str,
        timeout: int = DEFAULT_TIMEOUT_SECONDS,
        show_progress: bool = True,
        chunk_size: int = 8192,
) -> str:
    """Download text content from a URL with optional progress bar.

    Args:
      url: The URL to download from.
      timeout: Request timeout in seconds.
      show_progress: Whether to show a progress bar during download.
      chunk_size: Size of chunks to download at a time.

    Returns:
      The text content of the URL.

    Raises:
      requests.RequestException: If the download fails.
      ValueError: If the content is not text-based.
    """
    try:
        # Make initial request to get headers
        response = requests.get(url, stream=True, timeout=timeout)
        response.raise_for_status()

        # Check content type
        content_type = response.headers.get('Content-Type', '').lower()
        if not any(
                ct in content_type
                for ct in ['text/', 'application/json', 'application/xml']
        ):
            # Try to proceed anyway, but warn
            print(f"Warning: Content-Type '{content_type}' may not be text-based")

        # Get content length for progress bar
        total_size = int(response.headers.get('Content-Length', 0))

        filename = url.split('/')[-1][:50]

        # Download content with progress bar
        chunks = []
        if show_progress and total_size > 0:
            progress_bar = progress.create_download_progress_bar(
                total_size=total_size, url=url
            )

            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    chunks.append(chunk)
                    progress_bar.update(len(chunk))

            progress_bar.close()
        else:
            # Download without progress bar
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    chunks.append(chunk)

        # Combine chunks and decode
        content = b''.join(chunks)

        # Try to decode as text
        encodings = ['utf-8', 'latin-1', 'ascii', 'utf-16']
        text_content = None
        for encoding in encodings:
            try:
                text_content = content.decode(encoding)
                break
            except UnicodeDecodeError:
                continue

        if text_content is None:
            raise ValueError(f'Could not decode content from {url} as text')

        # Show content summary with clean formatting
        if show_progress:
            char_count = len(text_content)
            word_count = len(text_content.split())
            progress.print_download_complete(char_count, word_count, filename)

        return text_content

    except requests.RequestException as e:
        raise requests.RequestException(
            f'Failed to download from {url}: {str(e)}'
        ) from e



import os
from typing import Dict, Any

from fastmcp import FastMCP, Context

from workspacex.artifact import ArtifactType
from workspacex.base import WorkspaceConfig
from workspacex.utils.common import load_workspace

mcp = FastMCP("WorkspaceX Local RAG Tools 🔍")

class MCPConfig:
    """MCP client configuration"""
    def __init__(
        self,
        workspace_config: WorkspaceConfig = None
    ):
        # Priority: constructor param > env var > default
        self.workspace_id = os.getenv("WORKSPACE_ID", "test")
        self.workspace_type = os.getenv("WORKSPACE_TYPE", "local")
        self.threshold = os.getenv("SEARCH_THRESHOLD", 0.8)
        self.filter_types = os.getenv("FILTER_TYPES", None)
        self.workspace_config = workspace_config or WorkspaceConfig()
        
        self._load_workspace()

    async def _load_workspace(self):
        self.workspace = await load_workspace(self.workspace_id, self.workspace_type, config=self.workspace_config)


@mcp.tool
async def rag_search(
    query: str,
    limit: int = 10,
    ctx: Context = None
) -> Dict[str, Any]:
    """
    Search artifacts in the workspace using semantic search.
    
    Args:
        query: The search query text
        limit: Maximum number of results to return (default: 10)
        threshold: Similarity threshold for results (default: 0.8)
        filter_types: Optional list of artifact types to filter (e.g. ["CODE", "NOVEL"])
        ctx: MCP context (automatically injected)
        
    Returns:
        Dict containing search results or error message
    """
    config = MCPConfig()
    
    try:
        # 使用Context进行HTTP请求
        await ctx.info(f"Searching for: {query}")
        
        # 转换过滤类型（如果提供）
        artifact_types = None
        if config.filter_types:
            artifact_types = [ArtifactType(t) for t in config.filter_types]
        
        results = await config.workspace.search_artifacts(query, limit, config.threshold, artifact_types)
        
        await ctx.info(f"Found {len(results.get('results', []))} results")
        return results
    
    except Exception as e:
        await ctx.error(f"Error during search: {str(e)}")
        return {"error": f"Search failed: {str(e)}"}

if __name__ == "__main__":
    # 运行MCP服务器
    mcp.run()



from typing import List

from langchain_core.documents import Document
from langchain_text_splitters import MarkdownHeaderTextSplitter

from workspacex.artifact import Artifact, ChunkMetadata
from workspacex.chunk.base import Chunk, ChunkConfig, ChunkerBase


class MarkdownChunker(ChunkerBase):
    """
    A chunker that splits markdown content by headers.
    """
    
    def __init__(self, config: ChunkConfig):
        super().__init__(config=config)
        # Define headers to split on: (header_level, header_name)
        headers_to_split_on = [
            ("#", "Header 1"),
            ("##", "Header 2"),
            ("###", "Header 3")
        ]
        self._text_splitter = MarkdownHeaderTextSplitter(headers_to_split_on=headers_to_split_on)
        
    async def chunk(self, artifact: Artifact) -> List[Chunk]:
        """
        Chunks the given markdown content by headers.
        """
        texts = self._text_splitter.split_text(artifact.content)
        return self._create_chunks(texts, artifact)

    def _create_chunks(self, texts: List[Document], artifact: Artifact) -> List[Chunk]:
        chunks: List[Chunk] = []
        for i, text in enumerate(texts):
            content = (f"Header#1 {text.metadata.get("Header 1")}\n"
                      f"Header#2: {text.metadata.get("Header 2")}\n"
                      f"Header#3: {text.metadata.get("Header 3")}\n"
                      f"Content: \n\n  {text.page_content}"
                       )
            chunk = Chunk(
                chunk_id=f"{artifact.artifact_id}_chunk_{i}",
                content=content,
                chunk_metadata=ChunkMetadata(
                    chunk_index=i,
                    chunk_size=len(text.page_content),
                    artifact_id=artifact.artifact_id,
                    artifact_type=artifact.artifact_type.value,
                    parent_artifact_id=artifact.parent_id,
                )
            )
            chunks.append(chunk)
        return chunks



import uuid
from datetime import datetime
from enum import Enum, auto
from typing import Dict, Any, Optional, List

from pydantic import BaseModel, Field, model_validator


class ArtifactType(Enum):
    """Defines supported artifact types"""
    TEXT = "TEXT"
    CODE = "CODE"
    MARKDOWN = "MARKDOWN"
    HTML = "HTML"
    SVG = "SVG"
    JSON = "JSON"
    CSV = "CSV"
    TABLE = "TABLE"
    CHART = "CHART"
    DIAGRAM = "DIAGRAM"
    MCP_CALL = "MCP_CALL"
    TOOL_CALL = "TOOL_CALL"
    LLM_OUTPUT = "LLM_OUTPUT"
    WEB_PAGES = "WEB_PAGES"
    DIR = "DIR"
    CUSTOM = "CUSTOM"
    NOVEL = "NOVEL"
    CHUNK = "CHUNK"
    NOVEL_CHAPTER = "NOVEL_CHAPTER"
    ARXIV = "ARXIV"
    PDF = "PDF"


class ChunkMetadata(BaseModel):
    chunk_index: int = Field(default=0, description="Chunk index")
    chunk_size: int = Field(default=0, description="Chunk size")
    chunk_overlap: int = Field(default=0, description="Chunk overlap")
    artifact_id: str = Field(default="", description="Origin artifact ID")
    artifact_type: str = Field(default="", description="Origin artifact type")
    parent_artifact_id: str = Field(default=None, description="Parent artifact ID")

class Chunk(BaseModel):
    chunk_id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="Chunk ID")
    chunk_metadata: ChunkMetadata = Field(default=ChunkMetadata(), description="Chunk metadata")
    content: str = Field(default="", description="Chunk content")

    @property
    def parent_artifact_id(self) -> str:
        return self.chunk_metadata.parent_artifact_id
    
    @property
    def artifact_id(self) -> str:
        return self.chunk_metadata.artifact_id
    
    @property
    def artifact_type(self) -> str:
        return self.chunk_metadata.artifact_type
    
    @property
    def chunk_file_name(self) -> str:
        return f"{self.artifact_id}_chunk_{self.chunk_metadata.chunk_index}.json"

    def pre_n_chunk_file_name(self, pre_n) -> str:
        return f"{self.artifact_id}_chunk_{self.chunk_metadata.chunk_index - pre_n}.json"


    def next_n_chunk_file_name(self, next_n) -> str:
        return f"{self.artifact_id}_chunk_{self.chunk_metadata.chunk_index + next_n}.json"

class ArtifactStatus(Enum):
    """Artifact status"""
    DRAFT = auto()      # Draft status
    COMPLETE = auto()   # Completed status
    EDITED = auto()     # Edited status
    ARCHIVED = auto()   # Archived status

class AttachmentFile(BaseModel):
    file_id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="File ID")
    file_name: str = Field(default=None, description="File name")
    file_desc: str = Field(default_factory=lambda: str(uuid.uuid4()), description="File description")
    file_path: str = Field(default_factory=lambda: str(uuid.uuid4()), description="Temp File path", exclude=True)

class Artifact(BaseModel):
    """
    Represents a specific content generation result (artifact)
    
    Artifacts are the basic units of Artifacts technology, representing a structured content unit
    Can be code, markdown, charts, and various other formats
    """

    artifact_id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="Unique identifier for the artifact")
    parent_id: str = Field(default="", description="Parent identifier for the artifact")
    artifact_type: ArtifactType = Field(..., description="Type of the artifact")
    content: Any = Field(..., description="Content of the artifact")
    metadata: Dict[str, Any] = Field(default_factory=dict, description="Metadata associated with the artifact")
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Creation timestamp")
    updated_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Last updated timestamp")
    status: ArtifactStatus = Field(default=ArtifactStatus.ARCHIVED, description="Current status of the artifact")
    current_version: str = Field(default="", description="Current version of the artifact")
    version_history: list = Field(default_factory=list, description="History of versions for the artifact")
    create_file: bool = Field(default=False, description="Flag to indicate if a file should be created")
    sublist: List['Artifact'] = Field(default_factory=list, description="List of sub-artifacts (children)")
    chunk_list: List[Chunk] = Field(default_factory=list, description="List of chunks")
    kv_store: Dict[str, Any] = Field(default_factory=dict, description="key-value store do not persist", exclude=True)

    # Use model_validator for initialization logic
    @model_validator(mode='after')
    def setup_artifact(self):
        """Initialize the artifact after validation"""
        # Ensure artifact_id is always a valid string
        if not self.artifact_id:
            self.artifact_id = str(uuid.uuid4())
            
        # Reset status to DRAFT for new artifacts
        if not self.version_history:
            self.status = ArtifactStatus.DRAFT
            # Record initial version
            self._record_version("Initial version")
        
        return self

    def _record_version(self, description: str) -> None:
        """Record current state as a new version"""
        version = {
            "timestamp": datetime.now().isoformat(),
            "description": description,
            "content": self.content,
            "status": self.status
        }
        self.version_history.append(version)
        self.updated_at = version["timestamp"]

    def update_content(self, content: Any, description: str = "Content update") -> None:
        """
        Update artifact content and record version
        
        Args:
            content: New content
            description: Update description
        """
        self.content = content
        self.status = ArtifactStatus.EDITED
        self._record_version(description)

    def update_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Update artifact metadata
        
        Args:
            metadata: New metadata (will be merged with existing metadata)
        """
        self.metadata.update(metadata)
        self.updated_at = datetime.now().isoformat()

    def mark_complete(self) -> None:
        """Mark the artifact as complete"""
        self.status = ArtifactStatus.COMPLETE
        self._record_version("Marked as complete")

    def mark_chunkable(self) -> None:
        self.metadata['chunkable'] = True

    def archive(self) -> None:
        """Archive the artifact"""
        self.status = ArtifactStatus.ARCHIVED
        self._record_version("Artifact archived")

    def get_version(self, index: int) -> Optional[Dict[str, Any]]:
        """Get version at the specified index"""
        if 0 <= index < len(self.version_history):
            return self.version_history[index]
        return None

    def revert_to_version(self, index: int) -> bool:
        """Revert to a specific version"""
        version = self.get_version(index)
        if version:
            self.content = version["content"]
            self.status = version["status"]
            self._record_version(f"Reverted to version {index}")
            return True
        return False

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert artifact to dictionary, including sublist.
        Returns:
            Dictionary representation of the artifact
        """
        return {
            "artifact_id": self.artifact_id,
            "artifact_type": self.artifact_type.value,
            "content": self.content,
            "metadata": self.metadata,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "status": self.status.name,
            "parent_id": self.parent_id,
            # "version_count": len(self.version_history),
            # "version_history": self.version_history,
            "sublist": [sub.to_dict() for sub in self.sublist] if self.sublist else []
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Optional["Artifact"]:
        """
        Create an artifact instance from a dictionary, including sublist.
        Args:
            data: Dictionary data
        Returns:
            Artifact instance or None if artifact_id is not present
        """
        if not data.get("artifact_id"):
            return None
        artifact_type = ArtifactType(data.get("artifact_type"))
        artifact = cls(
            artifact_type=artifact_type,
            content=data.get("content"),
            metadata=data.get("metadata"),
            artifact_id=data.get("artifact_id")
        )
        artifact.parent_id = data.get("parent_id")
        artifact.created_at = data.get("created_at")
        artifact.updated_at = data.get("updated_at")
        artifact.status = ArtifactStatus[data.get("status")]
        # If version history exists, restore it as well
        if "version_history" in data:
            artifact.version_history = data.get("version_history")
        # Restore sublist if present
        if "sublist" in data and data.get("sublist"):
            artifact.sublist = [cls.from_dict(sub) for sub in data.get("sublist")]
        return artifact

    def add_subartifact(self, subartifact: 'Artifact') -> None:
        """
        Add a subartifact (child artifact) to this artifact.
        Args:
            subartifact: The subartifact to add
        """
        self.sublist.append(subartifact)

    def get_embedding_text(self) -> Optional[str]:
        """
        Get the embedding text for the artifact.
        """
        if not self.content:
            return None
        return str(self.content)

    def get_reranked_text(self) -> Optional[str]:
        """
        Get the reranked text for the artifact.
        """
        if not self.content:
            return None
        return str(self.content)

    def get_chunk_list(self) -> list[Chunk]:
        return self.chunk_list

    def get_metadata_value(self, key: str) -> Any:
        """
        Get the value of a metadata field by key.
        Args:
            key (str): The metadata key to retrieve
        Returns:
            Any: The value associated with the key, or None if not found
        """
        return self.metadata.get(key)

    @property
    def support_chunking(self):
        return True

    @property
    def summary(self):
        return self.metadata.get('summary')

    @summary.setter
    def summary(self, summary: str):
        self.metadata['summary'] = summary

    @property
    def attachment_files(self) -> Dict[str, AttachmentFile]:
        """Get attachment files dictionary
        
        Returns:
            Dict[str, AttachmentFile]: Dictionary of attachment files
        """
        if 'attachment_files' not in self.metadata:
            self.metadata['attachment_files'] = {}
        return self.metadata['attachment_files']
    
    def attachment_files_desc(self) -> Dict[str, str]:
        """
        desc for LLM to use
        """
        return {file.file_id: file.file_desc for file in self.attachment_files.values()}
    
    def get_attachment_file(self, file_id: str) -> Optional[AttachmentFile]:
        return self.attachment_files.get(file_id)

    def add_attachment_file(self, file: AttachmentFile):
        """Add an attachment file to the artifact
        
        Args:
            file (AttachmentFile): The attachment file to add
        """
        if 'attachment_files' not in self.metadata:
            self.metadata['attachment_files'] = {}
        self.metadata['attachment_files'][file.file_id] = file

    def after_chunker(self):
        pass


class SummaryArtifact(Artifact):

    origin_artifact: Artifact = Field(default=None)

    def __init__(self, origin_artifact: Artifact, **kwargs):
        if kwargs and isinstance(kwargs, dict):
            if kwargs.get("artifact_type"):
                kwargs.pop("artifact_type")
        super().__init__(artifact_type=ArtifactType.TEXT, **kwargs)
        self.origin_artifact = origin_artifact


class HybridSearchQuery(BaseModel):
    query: str = Field(..., description="Query string")
    filter_types: Optional[List[ArtifactType]] = Field(default=None, description="Filter types")
    limit: int = Field(default=3, description="Limit")
    threshold: float = Field(default=0.8, description="Threshold")

class HybridSearchResult(BaseModel):
    """
    Search result for an artifact
    """
    artifact: Artifact = Field(..., description="Artifact")
    score: float = Field(..., description="Score")
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Creation timestamp")

class ChunkSearchQuery(BaseModel):
    query: str = Field(..., description="Query string")
    limit: int = Field(default=10, description="Limit")
    threshold: float = Field(default=0.8, description="Threshold")
    pre_n: int = Field(default=3, description="Pre n")
    next_n: int = Field(default=3, description="Next n")
    filters: dict = Field(default_factory=dict, description="filters")

class ChunkSearchResult(BaseModel):
    """
    Search result for a chunk
    """
    chunk: Chunk = Field(..., description="Chunk")
    pre_n_chunks: Optional[List[Chunk]] = Field(default_factory=list, description="Pre n chunk")
    next_n_chunks: Optional[List[Chunk]] = Field(default_factory=list, description="Next n chunk")
    score: float = Field(..., description="relevant Score")
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Creation timestamp")



import uuid
from typing import Dict, Any, Optional

from pydantic import Field

from workspacex.artifact import Artifact, ArtifactType
from workspacex.extractor import noval_extractor


class NovelArtifact(Artifact):
    """
    Artifact for handling novels. Loads a novel file, splits it by chapters, and uploads each chapter as a separate file in the artifact's folder.
    """
    novel_file_path: str = Field(..., description="Path to the novel file")
    chapter_num: int = Field(default=0, description="Number of chapters")

    def __init__(
        self,
        artifact_type: ArtifactType,
        novel_file_path: str,
        metadata: Optional[Dict[str, Any]] = None,
        artifact_id: Optional[str] = None,
        **kwargs
    ):
        """
        Initialize a NovelArtifact, load and split the novel file, and create subartifacts for each chapter.
        """
        super().__init__(
            artifact_id=artifact_id or str(uuid.uuid4()),
            artifact_type=artifact_type,
            content=None,
            metadata=metadata or {},
            novel_file_path=novel_file_path,
            **kwargs
        )
        # Create subartifacts for each chapter
        artifacts = noval_extractor.extract(novel_file_path)
        self.chapter_num = len(artifacts)
        for sub_artifact in artifacts:
            sub_artifact.parent_id = self.artifact_id
            sub_artifact.mark_complete()
            self.add_subartifact(sub_artifact)
        self.mark_complete()

    @property
    def support_chunking(self):
        return False




import math
import re
from collections import Counter
from typing import List, Optional

from workspacex.artifact import Artifact
from workspacex.reranker.base import BaseRerankRunner, RerankConfig, RerankResult
from workspacex.utils.logger import logger
from workspacex.utils.timeit import timeit


class BM25RerankRunner(BaseRerankRunner):
    """
    BM25 Rerank Runner implementation.
    BM25 is a ranking function used by search engines to rank documents based on relevance.
    """

    def __init__(self, config: RerankConfig) -> None:
        """
        Initialize BM25RerankRunner.
        Args:
            config (RerankConfig): Configuration for BM25 parameters.
        """
        self.config = config
        # BM25 parameters with defaults
        self.k1 = getattr(config, 'k1', 1.2)  # Term frequency saturation parameter
        self.b = getattr(config, 'b', 0.75)   # Length normalization parameter
        self.avgdl = 0.0  # Average document length
        self.doc_freq = {}  # Document frequency for each term
        self.doc_lengths = {}  # Document lengths
        self.term_freq = {}  # Term frequency in each document
        self.corpus_size = 0  # Total number of documents

    def run(
            self,
            query: str,
            documents: List[Artifact],
            score_threshold: Optional[float] = 0.8,
            top_n: Optional[int] = 5,
            user: Optional[str] = None,
    ) -> List[RerankResult]:
        """
        Run BM25 rerank model.
        Args:
            query (str): Search query.
            documents (List[Artifact]): Documents for reranking.
            score_threshold (Optional[float]): Score threshold.
            top_n (Optional[int]): Top n results.
            user (Optional[str]): Unique user id if needed.
        Returns:
            List[RerankResult]: List of rerank results.
        """
        return self._run_bm25(query, documents, score_threshold, top_n, user)

    @timeit(logger.info, "BM25RerankRunner._run_bm25 took {elapsed_time:.3f} seconds")
    def _run_bm25(
            self,
            query: str,
            documents: List[Artifact],
            score_threshold: Optional[float] = 0.8,
            top_n: Optional[int] = 10,
            user: Optional[str] = None,
    ) -> List[RerankResult]:
        """
        Run BM25 rerank algorithm.
        """
        if not documents:
            return []

        # Preprocess documents and build corpus statistics
        self._build_corpus_stats(documents)
        
        # Tokenize query
        query_terms = self._tokenize(query)
        
        # Calculate BM25 scores for each document
        results = []
        for i, doc in enumerate(documents):
            score = self._calculate_bm25_score(query_terms, i)
            logger.info(f"📊 rerank_results item: {doc.artifact_id}: {score}")
            if score_threshold is None or score >= score_threshold:
                results.append(RerankResult(artifact=doc, score=score))
        
        # Sort by score in descending order
        results.sort(key=lambda x: x.score, reverse=True)
        
        # Apply top_n filter
        if top_n is not None:
            results = results[:top_n]
        
        return results

    def _build_corpus_stats(self, documents: List[Artifact]) -> None:
        """
        Build corpus statistics for BM25 calculation.
        Args:
            documents (List[Artifact]): List of documents to process.
        """
        self.corpus_size = len(documents)
        self.doc_lengths = {}
        self.term_freq = {}
        self.doc_freq = {}
        
        total_length = 0
        
        # Process each document
        for i, doc in enumerate(documents):
            text = doc.get_reranked_text()
            terms = self._tokenize(text)
            
            # Store document length
            doc_length = len(terms)
            self.doc_lengths[i] = doc_length
            total_length += doc_length
            
            # Count term frequencies in this document
            term_counts = Counter(terms)
            self.term_freq[i] = dict(term_counts)
            
            # Update document frequency for each unique term
            for term in term_counts:
                if term not in self.doc_freq:
                    self.doc_freq[term] = 0
                self.doc_freq[term] += 1
        
        # Calculate average document length
        self.avgdl = total_length / self.corpus_size if self.corpus_size > 0 else 0

    def _tokenize(self, text: str) -> List[str]:
        """
        Tokenize text into terms.
        Args:
            text (str): Text to tokenize.
        Returns:
            List[str]: List of tokens.
        """
        # Simple tokenization - split on whitespace and remove punctuation
        tokens = re.findall(r'\b\w+\b', text.lower())
        return tokens

    def _calculate_bm25_score(self, query_terms: List[str], doc_id: int) -> float:
        """
        Calculate BM25 score for a document given query terms.
        Args:
            query_terms (List[str]): Query terms.
            doc_id (int): Document ID.
        Returns:
            float: BM25 score.
        """
        score = 0.0
        
        for term in query_terms:
            # Skip terms not in vocabulary
            if term not in self.doc_freq:
                continue
                
            # Calculate IDF (Inverse Document Frequency)
            idf = math.log((self.corpus_size - self.doc_freq[term] + 0.5) / 
                          (self.doc_freq[term] + 0.5) + 1)
            
            # Get term frequency in this document
            tf = self.term_freq[doc_id].get(term, 0)
            
            # Calculate BM25 score component for this term
            doc_length = self.doc_lengths[doc_id]
            numerator = tf * (self.k1 + 1)
            denominator = tf + self.k1 * (1 - self.b + self.b * (doc_length / self.avgdl))
            
            if denominator > 0:
                score += idf * (numerator / denominator)
        
        return score



from abc import ABC, abstractmethod
from typing import Optional, List

from workspacex.embedding.base import EmbeddingsResult, EmbeddingsResults


class VectorDB(ABC):
    """Abstract base class for vector databases.
    
    This class defines the standard interface that all vector database implementations
    must follow. It provides methods for storing, retrieving, and searching vectors.
    """
    
    @abstractmethod
    def has_collection(self, collection_name: str) -> bool:
        """Check if a collection exists.
        
        Args:
            collection_name (str): Name of the collection
            
        Returns:
            bool: True if collection exists, False otherwise
        """
        pass
        
    @abstractmethod
    def delete_collection(self, collection_name: str):
        """Delete a collection.
        
        Args:
            collection_name (str): Name of the collection to delete
        """
        pass
        
    @abstractmethod
    def search(
        self, collection_name: str, vectors: list[list[float | int]], filter: dict, threshold: float,  limit: int
    ) -> Optional[EmbeddingsResults]:
        """Search for nearest neighbors based on vector similarity.
        
        Args:
            collection_name (str): Name of the collection
            vectors (list[list[float | int]]): Query vectors
            filter (dict): Filter conditions
            threshold (float): Threshold for similarity search
            limit (int): Maximum number of results to return
            
        Returns:
            Optional[EmbeddingsResults]: Search results or None if collection doesn't exist
        """
        pass
        
    @abstractmethod
    def query(
        self, collection_name: str, filter: dict, limit: Optional[int] = None
    ) -> Optional[EmbeddingsResults]:
        """Query items from the collection based on filter.
        
        Args:
            collection_name (str): Name of the collection
            filter (dict): Filter conditions
            limit (Optional[int]): Maximum number of results to return
            
        Returns:
            Optional[EmbeddingsResults]: Query results or None if collection doesn't exist
        """
        pass
        
    @abstractmethod
    def get(self, collection_name: str) -> Optional[EmbeddingsResults]:
        """Get all items in the collection.
        
        Args:
            collection_name (str): Name of the collection
            
        Returns:
            Optional[EmbeddingsResults]: All items in the collection or None if collection doesn't exist
        """
        pass
        
    @abstractmethod
    def insert(self, collection_name: str, items: list[EmbeddingsResult]):
        """Insert items into the collection.
        
        Args:
            collection_name (str): Name of the collection
            items (list[EmbeddingsResult]): List of embedding results to insert
        """
        pass
        
    @abstractmethod
    def upsert(self, collection_name: str, items: list[EmbeddingsResult]):
        """Update or insert items in the collection.
        
        Args:
            collection_name (str): Name of the collection
            items (list[EmbeddingsResult]): List of embedding results to upsert
        """
        pass
        
    @abstractmethod
    def delete(
        self,
        collection_name: str,
        ids: Optional[list[str]] = None,
        filter: Optional[dict] = None,
    ):
        """Delete items from the collection.
        
        Args:
            collection_name (str): Name of the collection
            ids (Optional[list[str]]): List of item IDs to delete
            filter (Optional[dict]): Filter conditions for items to delete
        """
        pass
        
    @abstractmethod
    def reset(self):
        """Reset the database.
        
        This will delete all collections and item entries.
        """
        pass 


import json
import logging
import traceback
import uuid
from abc import ABC, abstractmethod
from typing import Any, Optional

from workspacex.artifact import Artifact, ArtifactType
from workspacex.utils.rag_utils import call_llm, call_llm_async


class BaseExtractor(ABC):
    """Base class for extractors.
    Extractors are used to extract artifacts from content. such as text, images, audio, video, html, pdf, markdown, noval.txt, etc.
    """

    @abstractmethod
    def extract(self, content: Any) -> Optional[list[Artifact]]:
        """Extract artifacts from content."""
        raise NotImplementedError

    @abstractmethod
    async def async_extract(self, content: Any, **kwargs) -> Optional[list[Artifact]]:
        """Extract artifacts from content."""
        raise NotImplementedError


class BaseLLMExtractor(BaseExtractor):
    def __init__(self, extractor_model: str, extractor_prompt: str, extract_type: str, format_json: bool = False):
        self.extractor_model = extractor_model
        self.extractor_prompt = extractor_prompt
        self.format_json = format_json
        self.extract_type = extract_type

    def extract(self, content: Any, llm_config: dict[str, Any] = None) -> list[Artifact]:
        try:
            extracted_data = self.extract_by_llm(content, llm_config)
            artifact = Artifact(
                artifact_id=f"{str(uuid.uuid4().hex)}_{self.extract_type}",
                content=f"{extracted_data}",
                artifact_type=ArtifactType.TEXT
            )
            return [artifact]
        except Exception as e:
            logging.error(f"[Extractor#{self.extract_type}] error: {e}, \n trace is {traceback.format_exc()}")
            raise e

    def extract_by_llm(self, content: Any, llm_config: dict[str, Any] = None) -> str:
        prompt = self.extractor_prompt.format(text=content)
        extract_model = self.extractor_model
        if llm_config and llm_config.get("extract_model"):
            extract_model = llm_config.get("extract_model")
        llm_result = call_llm(prompt, extract_model, llm_config)
        if self.format_json:
            extracted_data = json.loads(llm_result.replace("```json", "").replace("```", ""))
            logging.debug(f"extracted_data: {extracted_data}")
            return extracted_data
        else:
            return llm_result

    async def async_extract(self, content: Any, llm_config: dict[str, Any] = None) -> list[Artifact]:
        try:
            extracted_data = await self.extract_by_llm_async(content, llm_config)
            artifact = Artifact(
                artifact_id=f"{str(uuid.uuid4().hex)}_{self.extract_type}",
                content=f"{extracted_data}",
                artifact_type=ArtifactType.TEXT
            )
            return [artifact]
        except Exception as e:
            logging.error(f"[Extractor#{self.extract_type}] error: {e}, \n trace is {traceback.format_exc()}")
            raise e

    async def extract_by_llm_async(self, content: Any, llm_config: dict[str, Any] = None) -> str:
        prompt = self.extractor_prompt.format(text=content)
        extract_model = self.extractor_model
        if llm_config and llm_config.get("extract_model"):
            extract_model = llm_config.get("extract_model")
        llm_result = await call_llm_async(prompt, extract_model, llm_config)
        if self.format_json:
            extracted_data = json.loads(llm_result.replace("```json", "").replace("```", ""))
            logging.debug(f"extracted_data: {extracted_data}")
            return extracted_data
        else:
            return llm_result


class BaseJudgeLLMExtractor(BaseLLMExtractor):

    def __init__(self, extractor_type: str,
                 extractor_model: str,
                 extractor_prompt: str,
                 format_json: bool,
                 judge_prompt: str,
                 judge_model: str,
                 format_judge_json: bool= False):
        super().__init__(extractor_model, extractor_prompt, extractor_type, format_json=format_json)
        self.judge_model = judge_model
        self.judge_prompt = judge_prompt
        self.format_judge_json = format_judge_json

    def extract(self, content: Any, llm_config: dict[str, Any] = None) -> list[Artifact]:
        artifacts = []
        try:
            extracted_data = self.extract_by_llm(content, llm_config)
            logging.info(f"Extractor#{self.extract_type} extracted_data: {extracted_data}")
            judge_data = self.judge(extracted_data)
            logging.info(f"Extractor#{self.extract_type} judge_data: {judge_data}")
            artifact = Artifact(
                artifact_id=f"{str(uuid.uuid4().hex)}_level_up",
                content=f"{judge_data}",
                artifact_type=ArtifactType.TEXT,
                metadata={
                    "origin_extracted_data": extracted_data,
                }
            )
            artifacts.append(artifact)
            return artifacts
        except Exception as e:
            logging.error(f"Error is {e}, \n trace is {traceback.format_exc()}")
            raise

    async def async_extract(self, content: Any, llm_config: dict[str, Any] = None) -> list[Artifact]:
        try:
            extracted_data = await self.extract_by_llm_async(content, llm_config)
            judge_data = await self.async_judge(extracted_data)
            artifact = Artifact(
                artifact_id=f"{str(uuid.uuid4().hex)}_level_up",
                content=f"{judge_data}",
                artifact_type=ArtifactType.TEXT,
                metadata={
                    "origin_extracted_data": extracted_data,
                }
            )
            return [artifact]
        except Exception as e:
            logging.error(f"[Extractor#{self.extract_type}] error: {e}, \n trace is {traceback.format_exc()}")
            raise

    @abstractmethod
    def judge(self, extracted_data: Any) -> Any:
        raise NotImplementedError
    @abstractmethod
    async def async_judge(self, extracted_data: Any) -> Any:
        raise NotImplementedError

    def _judge_by_llm(self, judge_prompt: str, llm_config: dict[str, Any] = None) -> Any:
        judge_result = call_llm(judge_prompt, self.judge_model, llm_config)
        if self.format_judge_json:
            judge_data = json.loads(judge_result.replace("```json", "").replace("```", ""))
            logging.debug(f"judge_data: {judge_data}")
            return judge_data
        else:
            return judge_result

    async def _async_judge_by_llm(self, judge_prompt: str, llm_config: dict[str, Any] = None) -> Any:
        judge_result = await call_llm_async(judge_prompt, self.judge_model, llm_config)
        if self.format_judge_json:
            judge_data = json.loads(judge_result.replace("```json", "").replace("```", ""))
            logging.debug(f"judge_data: {judge_data}")
            return judge_data
        else:
            return judge_result



"""
Full-text search module for workspacex.

This module provides full-text search capabilities for artifacts,
supporting various search engines like Elasticsearch.
"""

from .dbs.base import FulltextDB
from .factory import FulltextDBFactory, FulltextDBConfig

__all__ = ["FulltextDBFactory", "FulltextDBConfig", "FulltextDB"] 


import time
import traceback
from typing import Optional

import chromadb
from chromadb import Settings
from chromadb.utils.batch_utils import create_batches
from tqdm import tqdm

from workspacex.embedding.base import EmbeddingsResult, EmbeddingsResults
from workspacex.utils.logger import logger
from workspacex.vector.dbs.base import VectorDB


class ChromaVectorDB(VectorDB):
    """ChromaDB implementation of the VectorDB interface."""
    def __init__(self, config: dict = None):
        """
        Initialize ChromaDB client with configuration.
        
        Args:
            config (dict): Configuration dictionary containing ChromaDB settings
        """
        if config is None:
            config = {}
            
        # Default settings
        settings_dict = {
            "allow_reset": config.get("allow_reset", True),
            "anonymized_telemetry": config.get("anonymized_telemetry", False),
        }
        
        # Add auth settings if provided
        if config.get("chroma_client_auth_provider") is not None:
            settings_dict["chroma_client_auth_provider"] = config["chroma_client_auth_provider"]
        if config.get("chroma_client_auth_credentials") is not None:
            settings_dict["chroma_client_auth_credentials"] = config["chroma_client_auth_credentials"]

        # Use HTTP client if host is specified, otherwise use persistent client
        if config.get("http_host"):
            self.client = chromadb.HttpClient(
                host=config["http_host"],
                port=config.get("http_port", 8000),
                headers=config.get("http_headers", {}),
                ssl=config.get("http_ssl", False),
                tenant=config.get("tenant"),
                database=config.get("database"),
                settings=Settings(**settings_dict),
            )
        else:
            self.client = chromadb.PersistentClient(
                path=config.get("data_path", "./chroma_db"),
                settings=Settings(**settings_dict),
                tenant=config.get("tenant"),
                database=config.get("database"),
            )

    def has_collection(self, collection_name: str) -> bool:
        # Check if the collection exists based on the collection name.
        collection_names = [collection.name for collection in self.client.list_collections()]
        return collection_name in collection_names

    def delete_collection(self, collection_name: str):
        # Delete the collection based on the collection name.
        return self.client.delete_collection(name=collection_name)

    def search(
        self, collection_name: str, vectors: list[list[float | int]], filter: dict, threshold: float, limit: int
    ) -> Optional[EmbeddingsResults]:
        """Search for nearest neighbors based on vector similarity.
        
        Args:
            collection_name (str): Name of the collection
            vectors (list[list[float | int]]): Query vectors
            limit (int): Maximum number of results to return
            
        Returns:
            Optional[EmbeddingsResults]: Search results or None if collection doesn't exist
        """
        try:
            collection = self.client.get_collection(name=collection_name)
            if collection:
                # Convert simple key-value filters to ChromaDB operator format
                where_conditions = []
                if filter:
                    for key, value in filter.items():
                        where_conditions.append({key: {"$eq": value}})
                    where_filter = {"$and": where_conditions} if len(where_conditions) > 1 else where_conditions[0]
                else:
                    where_filter = None

                result = collection.query(
                    query_embeddings=vectors,
                    where=where_filter,
                    n_results=limit,
                )

                # chromadb has cosine distance, 2 (worst) -> 0 (best). Re-ordering to 0 -> 1
                # https://docs.trychroma.com/docs/collections/configure cosine equation
                distances: list = result["distances"][0]
                distances = [2 - dist for dist in distances]
                distances = [[dist / 2 for dist in distances]]

                docs = self._convert2_embedding_result_with_score(result= result, distances=distances, threshold=threshold)

                return EmbeddingsResults(
                    **{
                        "docs": docs,
                        "retrieved_at": int(time.time()),
                    }
                )
            return None
        except Exception as e:
            traceback.print_exc()
            logger.error(f"Error in search: {e}")
            return None

    def query(
        self, collection_name: str, filter: dict, limit: Optional[int] = None
    ) -> Optional[EmbeddingsResults]:
        """Query items from the collection based on filter.
        
        Args:
            collection_name (str): Name of the collection
            filter (dict): Filter conditions
            limit (Optional[int]): Maximum number of results to return
            
        Returns:
            Optional[EmbeddingsResults]: Query results or None if collection doesn't exist
        """
        try:
            collection = self.client.get_collection(name=collection_name)
            if collection:
                result = collection.get(
                    where=filter,
                    limit=limit,
                )

                docs = self._convert2EmbeddingResult(result)

                return EmbeddingsResults(
                    **{
                        "docs": docs,
                        "retrieved_at": int(time.time()),
                    }
                )
            return None
        except:
            return None

    def get(self, collection_name: str) -> Optional[EmbeddingsResults]:
        """Get all items in the collection.
        
        Args:
            collection_name (str): Name of the collection
            
        Returns:
            Optional[EmbeddingsResults]: All items in the collection or None if collection doesn't exist
        """
        collection = self.client.get_collection(name=collection_name)
        if collection:
            result = collection.get()
            docs = self._convert2EmbeddingResult(result)
            return EmbeddingsResults(
                **{
                    "docs": docs,
                    "retrieved_at": int(time.time()),
                }
            )
        return None
    
    def _convert2_embedding_result_with_score(self, result, distances=None, threshold=None):
        """Convert ChromaDB result to list of EmbeddingsResult.
        
        Args:
            result (dict): ChromaDB query result containing documents, metadatas and ids
            distances (Optional[List[List[float]]]): Similarity scores from search results
            
        Returns:
            list[EmbeddingsResult]: List of embedding results with content and metadata
        """
        from workspacex.embedding.base import EmbeddingsMetadata
        
        docs = []
        # Flatten distances if provided (ChromaDB returns nested list)
        scores = distances[0] if distances else [None] * len(result.get("documents", []))
        
        # ChromaDB returns nested lists for all fields
        documents = result.get("documents", [[]])[0]
        metadatas = result.get("metadatas", [[]])[0]
        ids = result.get("ids", [[]])[0]
        
        for document, metadata, id, score in zip(
            documents,
            metadatas,
            ids,
            scores
        ):
            # Metadata is already a dict since we stored it that way
            metadata_obj = EmbeddingsMetadata.model_validate(metadata)
            if threshold and score < threshold:
                continue
            
            docs.append(
                EmbeddingsResult(
                    id=id,
                    embedding=None,  # We don't need embeddings for retrieved results
                    content=document,
                    metadata=metadata_obj,
                    score=score
                )
            )
        return docs


    def _convert2EmbeddingResult(self, result):
        """Convert ChromaDB result to list of EmbeddingsResult.
        
        Args:
            result (dict): ChromaDB query result containing documents, metadatas and ids
            distances (Optional[List[List[float]]]): Similarity scores from search results
            
        Returns:
            list[EmbeddingsResult]: List of embedding results with content and metadata
        """
        from workspacex.embedding.base import EmbeddingsMetadata
        
        docs = []
        
        # ChromaDB returns nested lists for all fields
        documents = result.get("documents", [])
        metadatas = result.get("metadatas", [])
        ids = result.get("ids", [])
        
        for document, metadata, id in zip(
            documents,
            metadatas,
            ids
        ):
            # Metadata is already a dict since we stored it that way
            metadata_obj = EmbeddingsMetadata.model_validate(metadata)
            
            docs.append(
                EmbeddingsResult(
                    id=id,
                    embedding=None,  # We don't need embeddings for retrieved results
                    content=document,
                    metadata=metadata_obj,
                    score=None
                )
            )
        return docs

    def insert(self, collection_name: str, items: list[EmbeddingsResult]):
        """Insert the items into the collection.
        
        Args:
            collection_name (str): Name of the collection
            items (list[EmbeddingsResult]): List of embedding results to insert
        """
        collection = self.client.get_or_create_collection(
            name=collection_name, metadata={"hnsw:space": "cosine"}
        )

        ids = [item.id for item in items]
        documents = [item.content for item in items]
        embeddings = [item.embedding for item in items]
        # Convert metadata to dict instead of JSON string
        metadatas = [item.metadata.model_dump() for item in items]

        for batch in tqdm(create_batches(
            api=self.client,
            documents=documents,
            embeddings=embeddings,
            ids=ids,
            metadatas=metadatas,
        ), desc="save embeddings", ):
            collection.add(*batch)

    def upsert(self, collection_name: str, items: list[EmbeddingsResult]):
        """Update or insert items in the collection.
        
        Args:
            collection_name (str): Name of the collection
            items (list[EmbeddingsResult]): List of embedding results to upsert
        """
        collection = self.client.get_or_create_collection(
            name=collection_name, metadata={"hnsw:space": "cosine"}
        )

        ids = [item.id for item in items]
        documents = [item.content for item in items]
        embeddings = [item.embedding for item in items]
        # Convert metadata to dict instead of JSON string
        metadatas = [item.metadata.model_dump() for item in items]

        collection.upsert(
            ids=ids, documents=documents, embeddings=embeddings, metadatas=metadatas
        )

    def delete(
        self,
        collection_name: str,
        ids: Optional[list[str]] = None,
        filter: Optional[dict] = None,
    ):
        # Delete the items from the collection based on the ids.
        try:
            collection = self.client.get_collection(name=collection_name)
            if collection:
                if ids:
                    collection.delete(ids=ids)
                elif filter:
                    where_conditions = []
                    if filter:
                        for key, value in filter.items():
                            where_conditions.append({key: {"$eq": value}})
                        where_filter = {"$and": where_conditions} if len(where_conditions) > 1 else where_conditions[0]
                    else:
                        where_filter = None
                    collection.delete(where=where_filter)
                else:
                    self.client.delete_collection(name=collection_name)
        except Exception as e:
            # If collection doesn't exist, that's fine - nothing to delete
            logger.debug(
                f"Attempted to delete from non-existent collection {collection_name}. Ignoring."
            )
            pass

    def reset(self):
        # Resets the database. This will delete all collections and item entries.
        return self.client.reset()



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# reference https://github.com/google/langextract/blob/main/langextract/progress.py

"""Progress and visualization utilities for LangExtract."""

import urllib.parse
from typing import Any

import tqdm

from workspacex.utils.logger import logger

# ANSI color codes for terminal output
BLUE = "\033[94m"
GREEN = "\033[92m"
CYAN = "\033[96m"
BOLD = "\033[1m"
RESET = "\033[0m"

# Google Blue color for progress bars
GOOGLE_BLUE = "#4285F4"


def create_download_progress_bar(
    total_size: int, url: str, ncols: int = 100, max_url_length: int = 50
) -> tqdm.tqdm:
  """Create a styled progress bar for downloads.

  Args:
    total_size: Total size in bytes.
    url: The URL being downloaded.
    ncols: Number of columns for the progress bar.
    max_url_length: Maximum length to show for the URL.

  Returns:
    A configured tqdm progress bar.
  """
  # Truncate URL if too long, keeping the domain and end
  if len(url) > max_url_length:
    parsed = urllib.parse.urlparse(url)
    domain = parsed.netloc or parsed.hostname or "unknown"

    path_parts = parsed.path.strip("/").split("/")
    filename = path_parts[-1] if path_parts and path_parts[-1] else "file"

    available = max_url_length - len(domain) - len(filename) - 5
    if available > 0:
      url_display = f"{domain}/.../{filename}"
    else:
      url_display = url[: max_url_length - 3] + "..."
  else:
    url_display = url

  return tqdm.tqdm(
      total=total_size,
      unit="B",
      unit_scale=True,
      desc=(
          f"{BLUE}{BOLD}WORKSPACE{RESET}: Downloading"
          f" {GREEN}{url_display}{RESET}"
      ),
      bar_format=(
          "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}"
          " [{elapsed}<{remaining}, {rate_fmt}]"
      ),
      colour=GOOGLE_BLUE,
      ncols=ncols,
  )


def create_extraction_progress_bar(
    iterable: Any, model_info: str | None = None, disable: bool = False
) -> tqdm.tqdm:
  """Create a styled progress bar for extraction.

  Args:
    iterable: The iterable to wrap with progress bar.
    model_info: Optional model information to display (e.g., "gemini-1.5-pro").
    disable: Whether to disable the progress bar.

  Returns:
    A configured tqdm progress bar.
  """
  desc = format_extraction_progress(model_info)

  return tqdm.tqdm(
      iterable,
      desc=desc,
      bar_format="{desc} [{elapsed}]",
      disable=disable,
      dynamic_ncols=True,
  )


def print_download_complete(
    char_count: int, word_count: int, filename: str
) -> None:
  """Print a styled download completion message.

  Args:
    char_count: Number of characters downloaded.
    word_count: Number of words downloaded.
    filename: Name of the downloaded file.
  """
  logger.info(
      f"{GREEN}✓{RESET} Downloaded {BOLD}{char_count:,}{RESET} characters "
      f"({BOLD}{word_count:,}{RESET} words) from {BLUE}{filename}{RESET}",
      flush=True,
  )


def print_extraction_complete() -> None:
  """Print a generic extraction completion message."""
  logger.info(f"{GREEN}✓{RESET} Extraction processing complete", flush=True)


def print_extraction_summary(
    num_extractions: int,
    unique_classes: int,
    elapsed_time: float | None = None,
    chars_processed: int | None = None,
    num_chunks: int | None = None,
) -> None:
  """Print a styled extraction summary with optional performance metrics.

  Args:
    num_extractions: Total number of extractions.
    unique_classes: Number of unique extraction classes.
    elapsed_time: Optional elapsed time in seconds.
    chars_processed: Optional number of characters processed.
    num_chunks: Optional number of chunks processed.
  """
  logger.info(
      f"{GREEN}✓{RESET} Extracted {BOLD}{num_extractions}{RESET} entities "
      f"({BOLD}{unique_classes}{RESET} unique types)",
      flush=True,
  )

  if elapsed_time is not None:
    metrics = []

    # Time
    metrics.append(f"Time: {BOLD}{elapsed_time:.2f}s{RESET}")

    # Speed
    if chars_processed is not None and elapsed_time > 0:
      speed = chars_processed / elapsed_time
      metrics.append(f"Speed: {BOLD}{speed:,.0f}{RESET} chars/sec")

    if num_chunks is not None:
      metrics.append(f"Chunks: {BOLD}{num_chunks}{RESET}")

    for metric in metrics:
      logger.info(f"  {CYAN}•{RESET} {metric}", flush=True)


def create_save_progress_bar(
    output_path: str, disable: bool = False
) -> tqdm.tqdm:
  """Create a progress bar for saving documents.

  Args:
    output_path: The output file path.
    disable: Whether to disable the progress bar.

  Returns:
    A configured tqdm progress bar.
  """
  filename = output_path.split("/")[-1]
  return tqdm.tqdm(
      desc=(
          f"{BLUE}{BOLD}LangExtract{RESET}: Saving to {GREEN}{filename}{RESET}"
      ),
      unit=" docs",
      disable=disable,
  )


def create_load_progress_bar(
    file_path: str, total_size: int | None = None, disable: bool = False
) -> tqdm.tqdm:
  """Create a progress bar for loading documents.

  Args:
    file_path: The file path being loaded.
    total_size: Optional total file size in bytes.
    disable: Whether to disable the progress bar.

  Returns:
    A configured tqdm progress bar.
  """
  filename = file_path.split("/")[-1]
  if total_size:
    return tqdm.tqdm(
        total=total_size,
        desc=(
            f"{BLUE}{BOLD}LangExtract{RESET}: Loading {GREEN}{filename}{RESET}"
        ),
        unit="B",
        unit_scale=True,
        disable=disable,
    )
  else:
    return tqdm.tqdm(
        desc=(
            f"{BLUE}{BOLD}LangExtract{RESET}: Loading {GREEN}{filename}{RESET}"
        ),
        unit=" docs",
        disable=disable,
    )


def print_save_complete(num_docs: int, file_path: str) -> None:
  """Print a save completion message.

  Args:
    num_docs: Number of documents saved.
    file_path: Path to the saved file.
  """
  filename = file_path.split("/")[-1]
  logger.info(
      f"{GREEN}✓{RESET} Saved {BOLD}{num_docs}{RESET} documents to"
      f" {GREEN}{filename}{RESET}",
      flush=True,
  )


def print_load_complete(num_docs: int, file_path: str) -> None:
  """Print a load completion message.

  Args:
    num_docs: Number of documents loaded.
    file_path: Path to the loaded file.
  """
  filename = file_path.split("/")[-1]
  logger.info(
      f"{GREEN}✓{RESET} Loaded {BOLD}{num_docs}{RESET} documents from"
      f" {GREEN}{filename}{RESET}",
      flush=True,
  )


def get_model_info(language_model: Any) -> str | None:
  """Extract model information from a language model instance.

  Args:
    language_model: A language model instance.

  Returns:
    A string describing the model, or None if not available.
  """
  if hasattr(language_model, "model_id"):
    return language_model.model_id

  if hasattr(language_model, "model_url"):
    return language_model.model_url

  return None


def format_extraction_stats(current_chars: int, processed_chars: int) -> str:
  """Format extraction progress statistics with colors.

  Args:
    current_chars: Number of characters in current batch.
    processed_chars: Total number of characters processed so far.

  Returns:
    Formatted string with colored statistics.
  """
  current_str = f"{GREEN}{current_chars:,}{RESET}"
  processed_str = f"{GREEN}{processed_chars:,}{RESET}"
  return f"current={current_str} chars, processed={processed_str} chars"


def create_extraction_postfix(current_chars: int, processed_chars: int) -> str:
  """Create a formatted postfix string for extraction progress.

  Args:
    current_chars: Number of characters in current batch.
    processed_chars: Total number of characters processed so far.

  Returns:
    Formatted string with statistics.
  """
  current_str = f"{GREEN}{current_chars:,}{RESET}"
  processed_str = f"{GREEN}{processed_chars:,}{RESET}"
  return f"current={current_str} chars, processed={processed_str} chars"


def format_extraction_progress(
    model_info: str | None,
    current_chars: int | None = None,
    processed_chars: int | None = None,
) -> str:
  """Format the complete extraction progress bar description.

  Args:
    model_info: Optional model information (e.g., "gemini-2.0-flash").
    current_chars: Number of characters in current batch (optional).
    processed_chars: Total number of characters processed so far (optional).

  Returns:
    Formatted description string.
  """
  # Base description
  if model_info:
    desc = f"{BLUE}{BOLD}LangExtract{RESET}: model={GREEN}{model_info}{RESET}"
  else:
    desc = f"{BLUE}{BOLD}LangExtract{RESET}: Processing"

  # Add stats if provided
  if current_chars is not None and processed_chars is not None:
    current_str = f"{GREEN}{current_chars:,}{RESET}"
    processed_str = f"{GREEN}{processed_chars:,}{RESET}"
    desc += f", current={current_str} chars, processed={processed_str} chars"

  return desc


def create_pass_progress_bar(
    total_passes: int, disable: bool = False
) -> tqdm.tqdm:
  """Create a progress bar for sequential extraction passes.

  Args:
    total_passes: Total number of sequential passes.
    disable: Whether to disable the progress bar.

  Returns:
    A configured tqdm progress bar.
  """
  desc = f"{BLUE}{BOLD}LangExtract{RESET}: Extraction passes"
  return tqdm.tqdm(
      total=total_passes,
      desc=desc,
      bar_format=(
          "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}]"
      ),
      disable=disable,
      colour=GOOGLE_BLUE,
      ncols=100,
  )


import asyncio
import os
import traceback
import uuid
from datetime import datetime
from typing import Dict, Any, Optional, List, Union

from pydantic import BaseModel, Field, ConfigDict
from tqdm import tqdm

from workspacex.artifact import ArtifactType, Artifact, Chunk, ChunkSearchQuery, ChunkSearchResult, HybridSearchResult, \
    HybridSearchQuery
from workspacex.artifacts.arxiv import ArxivArtifact
from workspacex.base import WorkspaceConfig
from workspacex.chunk.base import ChunkerFactory
from workspacex.code_artifact import CodeArtifact
from workspacex.embedding.base import EmbeddingFactory
from workspacex.fulltext.dbs.base import FulltextDB, FulltextSearchResult
from workspacex.fulltext.factory import FulltextDBFactory
from workspacex.novel_artifact import NovelArtifact
from workspacex.observer import WorkspaceObserver, get_observer
from workspacex.reranker.base import RerankResult
from workspacex.reranker.factory import RerankerFactory
from workspacex.storage.base import BaseRepository
from workspacex.storage.local import LocalPathRepository
from workspacex.utils.logger import logger
from workspacex.vector.dbs.base import VectorDB
from workspacex.vector.factory import VectorDBFactory


class WorkSpace(BaseModel):
    """
    Artifact workspace, managing a group of related artifacts
    
    Provides collaborative editing features, supporting version management, update notifications, etc. for multiple Artifacts
    """

    workspace_id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="unique identifier for the workspace")
    name: str = Field(default="", description="name of the workspace")
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat())
    updated_at: str = Field(default_factory=lambda: datetime.now().isoformat())
    metadata: Dict[str, Any] = Field(default={}, description="metadata")
    artifacts: List[Artifact] = Field(default=[], description="list of artifacts")

    observers: Optional[List[WorkspaceObserver]] = Field(default=[], description="list of observers", exclude=True)
    repository: Optional[BaseRepository] = Field(default=None, description="local artifact repository", exclude=True)
    workspace_config: Optional[WorkspaceConfig] = Field(default=None, description="workspace config", exclude=True)
    
    vector_db: Optional[VectorDB] = Field(default=None, description="vector_db instance", exclude=True)
    fulltext_db: Optional[FulltextDB] = Field(default=None, description="fulltext_db instance", exclude=True)
    kv_db: Optional[dict[str, Any]] = Field(default_factory=dict, description="kv_db instance", exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True, extra='allow')
    
    def __init__(
            self,
            workspace_id: Optional[str] = None,
            name: Optional[str] = None,
            storage_path: Optional[str] = None,
            observers: Optional[List[WorkspaceObserver]] = None,
            use_default_observer: bool = True,
            clear_existing: bool = False,
            repository: Optional[BaseRepository] = None,
            **kwargs
    ):
        super().__init__()
        self.workspace_id = workspace_id or str(uuid.uuid4())
        self.name = name or f"Workspace-{self.workspace_id[:8]}"
        self.created_at = datetime.now().isoformat()
        self.updated_at = self.created_at
        if not kwargs.get("config") and not isinstance(kwargs.get("config"), WorkspaceConfig):
            self.workspace_config = WorkspaceConfig()
        else:
            self.workspace_config = kwargs["config"]

        # Initialize repository first
        if repository:
            self.repository = repository
        else:
            storage_dir = storage_path or os.path.join("data", "workspaces", self.workspace_id)
            self.repository = LocalPathRepository(storage_dir, clear_existing=clear_existing)

        # Initialize artifacts and metadata
        if clear_existing:
            self.artifacts = []
            self.metadata = {}
        else:
            # Try to load existing workspace data
            workspace_data = self._load_workspace_data()
            if workspace_data:
                self.artifacts = workspace_data.get('artifacts', [])
                self.metadata = workspace_data.get('metadata', {})
                self.created_at = workspace_data.get('created_at', self.created_at)
                self.updated_at = workspace_data.get('updated_at', self.updated_at)
            else:
                self.artifacts = []
                self.metadata = {}

        # Initialize observers
        self.observers: List[WorkspaceObserver] = []
        if use_default_observer:
            self.observers.append(get_observer())

        if observers:
            for observer in observers:
                if observer not in self.observers:  # Avoid duplicates
                    self.add_observer(observer)


        
        # Initialize full-text search database if enabled
        if self.workspace_config.fulltext_db_config and self.workspace_config.fulltext_db_config.provider:
            self.fulltext_db = FulltextDBFactory.get_fulltext_db(self.workspace_config.fulltext_db_config)
        else:
            self.fulltext_db = None
            
        # Initialize vector database if enabled
        if self.workspace_config.vector_db_config and self.workspace_config.vector_db_config.provider:
            self.vector_db = VectorDBFactory.get_vector_db(self.workspace_config.vector_db_config)
        else:
            self.vector_db = None
        
        self._reranker = None
        self._chunker = None
        self._embedder = None
        
        if clear_existing:
            if self.vector_db:
                self.vector_db.delete(self.default_vector_collection)
            if self.fulltext_db:
                self.fulltext_db.delete(self.workspace_id)

    @property
    def chunker(self):
        if not self._chunker:
            self._chunker = ChunkerFactory.get_chunker(self.workspace_config.chunk_config)
        return self._chunker

    @property
    def embedder(self):
        if not self._embedder:
            self._embedder = EmbeddingFactory.get_embedder(self.workspace_config.embedding_config)
        return self._embedder

    @property
    def reranker(self):
        if not self._reranker:
            self._reranker = RerankerFactory.getReranker(self.workspace_config.reranker_config)
        return self._reranker


    @property
    def default_vector_collection(self):
        return f"{self.workspace_id}"

    @property
    def summary_vector_collection(self):
        return f"summary_{self.workspace_id}"

    @property
    def full_text_index(self):
        return f"f_{self.workspace_id}"

    @classmethod
    def from_s3_storages(cls, workspace_id: Optional[str] = None,
                         name: Optional[str] = None,
                         storage_path: Optional[str] = None,
                         use_ssl=False,
                         **kwargs
                         ):
        from workspacex.storage.s3 import S3Repository

        # MinIO connection config
        s3_kwargs = {
            'key': os.getenv('MINIO_ACCESS_KEY'),
            'secret': os.getenv('MINIO_SECRET_KEY'),
            'client_kwargs': {
                'endpoint_url': os.getenv('MINIO_ENDPOINT_URL')
            },
            'use_ssl': False
        }
        bucket = os.getenv('MINIO_WORKSPACE_BUCKET')

        # Ensure bucket exists (create if not)
        import s3fs
        fs = s3fs.S3FileSystem(**s3_kwargs)
        if not fs.exists(bucket):
            fs.mkdir(bucket)

        # Create S3Repository
        repo = S3Repository(storage_path=storage_path,
                            bucket=bucket,
                            s3_kwargs=s3_kwargs)

        # Create a workspace using S3Repository
        return cls(workspace_id=workspace_id,
                   name=name,
                   repository=repo, **kwargs
                   )

    @classmethod
    def from_local_storages(cls, workspace_id: Optional[str] = None,
                            name: Optional[str] = None,
                            storage_path: Optional[str] = None,
                            observers: Optional[List[WorkspaceObserver]] = None,
                            use_default_observer: bool = True
                            , **kwargs

                            ) -> "WorkSpace":
        """
        Create a workspace instance from local storage
        
        Args:
            workspace_id: Optional workspace ID
            name: Optional workspace name
            storage_path: Optional storage path
            observers: Optional list of observers
            use_default_observer: Whether to use default observer
            
        Returns:
            WorkSpace instance
        """
        workspace = cls(
            workspace_id=workspace_id,
            name=name,
            storage_path=storage_path,
            observers=observers,
            use_default_observer=use_default_observer,
            clear_existing=False,  # Always try to load existing data
            **kwargs
        )
        return workspace

    #########################################################
    # Artifact Management
    #########################################################

    async def create_artifact(
            self,
            artifact_type: Union[ArtifactType, str],
            artifact_id: Optional[str] = None,
            content: Optional[Any] = None,
            metadata: Optional[Dict[str, Any]] = None,
            **kwargs
    ) -> List[Artifact]:
        """
        Create a new artifact
        Args:
            artifact_type: Artifact type (enum or string)
            artifact_id: Optional artifact ID (will be generated if not provided)
            content: Artifact content
            metadata: Metadata dictionary
            novel_file_path: Path to the novel file (for NOVEL type)
        Returns:
            List of created artifact objects
        """
        # If a string is passed, convert to enum type
        if isinstance(artifact_type, str):
            artifact_type = ArtifactType(artifact_type)

        # Create new artifacts
        artifacts = []
        artifact = None

        # Ensure metadata is a dictionary
        if metadata is None:
            metadata = {}
        
        # Ensure artifact_id is a valid string
        if artifact_id is None:
            artifact_id = str(uuid.uuid4())

        if artifact_type == ArtifactType.CODE:
            artifacts = CodeArtifact.from_code_content(artifact_type, content)
        elif artifact_type == ArtifactType.NOVEL:
            novel_file_path = kwargs.get('novel_file_path')
            if not novel_file_path:
                raise ValueError("novel_file_path must be provided for NOVEL artifact type")
            artifact = NovelArtifact(
                artifact_type=artifact_type,
                novel_file_path=novel_file_path,
                metadata=metadata,
                artifact_id=artifact_id
            )
        elif artifact_type == ArtifactType.ARXIV:
            if 'arxiv_id' not in kwargs:
                raise ValueError("novel_file_path must be provided for NOVEL artifact type")
            artifact = ArxivArtifact(
                arxiv_id=kwargs.get('arxiv_id'),
                page_count=kwargs.get('page_count', -1)
            )
            await artifact.process_arxiv()

        else:
            artifact = Artifact(
                artifact_id=artifact_id,
                artifact_type=artifact_type,
                content=content,
                metadata=metadata
            )
        if artifact:
            await self.add_artifact(artifact)
            return [artifact]
        
        if artifacts:
            for artifact in artifacts:
                await self.add_artifact(artifact)

        return artifacts

    async def add_artifact(
            self,
            artifact: Artifact
    ) -> None:
        """
        Create a new artifact

        Args:
            artifact: Artifact

        Returns:
            List of created artifact objects
        """
        # Check if artifact ID already exists
        existing_artifact = self._get_artifact(artifact.artifact_id)
        if existing_artifact:
            self._update_artifact(artifact)
            await self._notify_observers("update", artifact)
        else:
            # Add to workspace
            self.artifacts.append(artifact)

            await self._notify_observers("create", artifact)

        # Store in repository
        await self._store_artifact(artifact)

        # Update workspace time
        self.updated_at = datetime.now().isoformat()

        # Save workspace state to create new version
        self.save()


    def _update_artifact(self, artifact: Artifact) -> None:
        for i, a in enumerate(self.artifacts):
            if a.artifact_id == artifact.artifact_id:
                self.artifacts[i] = artifact
                logger.info(f"[📂WORKSPACEX]🔄 Updating artifact in repository: {artifact.artifact_id}")
                break

    async def update_artifact(
            self,
            artifact_id: str,
            content: Any,
            description: str = "Content update"
    ) -> Optional[Artifact]:
        """
        Update artifact content
        
        Args:
            artifact_id: Artifact ID
            content: New content
            description: Update description
            
        Returns:
            Updated artifact, or None if it doesn't exist
        """
        artifact = self._get_artifact(artifact_id)
        if artifact:
            artifact.update_content(content, description)

            # Update storage
            await self._store_artifact(artifact)

            # Update workspace time
            self.updated_at = datetime.now().isoformat()

            # Notify observers
            await self._notify_observers("update", artifact)

            return artifact
        return None
    
    async def update_artifact_metadata(self, artifact: Artifact, metadata: Dict[str, Any]) -> bool:
        """
        Update artifact metadata
        """
        if artifact.parent_id:
            parent_artifact = self._get_artifact(artifact.parent_id)
            if parent_artifact:
                for sub_artifact in parent_artifact.sublist:
                    if sub_artifact.artifact_id == artifact.artifact_id:
                        sub_artifact.update_metadata(metadata)
                        self.repository.store_artifact(parent_artifact, save_sub_list_content=False)
                        return True
        else:
            artifact.update_metadata(metadata)
            self.repository.store_artifact(artifact, save_sub_list_content=False)
        return True

    async def save_artifact(self, artifact: Artifact, save_sub_list_content=False):
        self.repository.store_artifact(artifact, save_sub_list_content=save_sub_list_content)

    async def delete_artifact(self, artifact_id: str) -> bool:
        """
        Delete an artifact from the workspace
        
        Args:
            artifact_id: Artifact ID
            
        Returns:
            Whether deletion was successful
        """
        for i, artifact in enumerate(self.artifacts):
            if artifact.artifact_id == artifact_id:
                # Mark as archived
                artifact.archive()
                # Store the archived state
                await self._store_artifact(artifact)
                # Remove from list
                self.artifacts.pop(i)

                # Update workspace time
                self.updated_at = datetime.now().isoformat()

                # Save workspace state to create new version
                self.save()

                # Notify observers
                await self._notify_observers("delete", artifact)
                return True
        return False
    
    async def _store_artifact(self, artifact: Artifact) -> None:
        """Store artifact in repository"""
        try:
            # Create semaphore to limit concurrent operations (default: 10)
            max_concurrent = getattr(self.workspace_config, 'max_concurrent_embeddings', 10)
            semaphore = asyncio.Semaphore(max_concurrent)

            # Process main artifact and subartifacts in parallel for maximum concurrency
            tasks = [self._store_artifact_chunk_and_embedding_with_semaphore(artifact, semaphore)]

            if artifact.sublist and len(artifact.sublist) > 0:
                # Add subartifacts to parallel processing
                sub_tasks = [self._store_artifact_chunk_and_embedding_with_semaphore(subartifact, semaphore)
                             for subartifact in artifact.sublist]
                tasks.extend(sub_tasks)
                logger.info(
                    f"🚀 Processing {len(tasks)} artifacts in parallel (1 main + {len(sub_tasks)} subartifacts) with max {max_concurrent} concurrent operations")

            # Execute all embedding tasks in parallel with error handling
            results = await asyncio.gather(*tasks, return_exceptions=True)

            # Log any errors that occurred during parallel processing
            errors = [r for r in results if isinstance(r, Exception)]
            if errors:
                logger.error(f"❌ {len(errors)} errors occurred during parallel :")
                for error in errors:
                    logger.error(f"[Workspace]- {type(error).__name__}: {error}")
            else:
                logger.info(f"✅ All {len(tasks)} artifacts processed successfully in parallel")
        finally:
            # save artifact
            self.repository.store_artifact(artifact=artifact)
            logger.info(f"📦[CONTENT] store_artifact[{artifact.artifact_type}]:{artifact.artifact_id} content finished")

    async def _store_artifact_chunk_and_embedding_with_semaphore(self, artifact: Artifact, semaphore: asyncio.Semaphore) -> None:
        """Store artifact embedding with concurrency control"""
        async with semaphore:
            return await self._chunk_and_embedding(artifact)

    async def _chunk_and_embedding(self, artifact: Artifact) -> None:
        """Store artifact embedding"""

        # if chunking is enabled, chunk the artifact first
        if self.workspace_config.chunk_config.enabled and artifact.support_chunking:
            chunks = await self._chunk_artifact(artifact)
            await self._rebuild_artifact_embedding(artifact, chunks)
            await self._rebuild_artifact_fulltext(artifact, chunks)
        else:
            await self._rebuild_artifact_embedding(artifact)
            await self._rebuild_artifact_fulltext(artifact)

    async def _chunk_artifact(self, artifact: Artifact) -> Optional[list[Chunk]]:
        """Chunk artifact"""
        chunker = self.get_chunker_by_artifact(artifact)
        if chunker:
            try:
                chunks = await chunker.chunk(artifact)
                artifact.mark_chunkable()
                logger.info(f"📦[CHUNKING]✅ store_artifact_chunk[{artifact.artifact_type}]:{artifact.artifact_id} chunks size: {len(chunks)}")
                
                if chunks:
                    await self.save_artifact_chunks(artifact, chunks)
                    artifact.after_chunk()
                    return chunks
                else:
                    logger.info(f"📦[EMBEDDING-CHUNKING]❌ store_artifact[{artifact.artifact_type}]:{artifact.artifact_id} chunks is empty")
                    return []
            except Exception as e:
                logger.error(f"📦[CHUNKING]❌ store_artifact[{artifact.artifact_type}]:{artifact.artifact_id} failed: {e}\n traceback is {traceback.format_exc()}")
                raise
        else:
            return []

    def get_chunker_by_artifact(self, artifact):
        if hasattr(artifact, "chunker"):
            return artifact.chunker
        return self.chunker

    async def rebuild_index(self):
        await self.rebuild_fulltext()
        await self.rebuild_embedding()

    async def rebuild_embedding(self):
        for artifact in tqdm(self.artifacts, desc="workspace_rebuild_embedding"):
            await self.rebuild_artifact_embedding(artifact)

    async def rebuild_artifact_embedding(self, artifact: Artifact, chunks: list[Chunk] = None):
        await self._rebuild_artifact_embedding(artifact)
        for sub_artifact in tqdm(artifact.sublist, f"artifact_rebuild_embedding_sublist#{artifact.artifact_id}"):
            await self._rebuild_artifact_embedding(sub_artifact)

    async def _rebuild_artifact_embedding(self, artifact: Artifact, chunks: list[Chunk] = None):
        if not self.workspace_config.embedding_config.enabled:
            return

        self.vector_db.delete(self.default_vector_collection, filter={"artifact_id": artifact.artifact_id})
        logger.info(f"📦[EMBEDDING]✅ delete_embeddings[{artifact.artifact_type}]:{artifact.artifact_id} finished")

        chunkable = artifact.get_metadata_value("chunkable")
        if chunkable:
            if not chunks:
                chunks = await self._load_artifact_chunks(artifact)
            if chunks:
                embedding_results = await self.embedder.async_embed_chunks(chunks)
                await asyncio.to_thread(self.vector_db.insert, self.default_vector_collection, embedding_results)
                logger.info(
                    f"📦[EMBEDDING-CHUNKING]✅ store_artifact[{artifact.artifact_type}]:{artifact.artifact_id} embedding_result finished")
        else:
            # if embedding is not enabled, skip
            if not artifact.get_embedding_text():
                logger.info(
                    f"📦[EMBEDDING]❌ store_artifact[{artifact.artifact_type}]:{artifact.artifact_id} embedding is not enabled or embedding_text is empty")
                return
            try:
                embedding_result = await asyncio.to_thread(self.embedder.embed_artifact, artifact)
                await asyncio.to_thread(self.vector_db.insert, self.default_vector_collection, [embedding_result])
                logger.info(
                    f"📦[EMBEDDING]✅ store_artifact(unchunkable)[{artifact.artifact_type}]:{artifact.artifact_id} embedding_result finished")
            except Exception as e:
                logger.error(
                    f"📦[EMBEDDING]❌ store_artifact(unchunkable)[{artifact.artifact_type}]:{artifact.artifact_id} failed: {e}")
                raise

    async def _load_artifact_chunks(self, artifact: Artifact) -> Optional[list[Chunk]]:
        return self.repository.get_chunks(artifact.artifact_id, artifact.parent_id)
    
    async def save_artifact_chunks(self, artifact: Artifact, chunks: List[Chunk]) -> None:
        """Save artifact chunks"""
        self.repository.store_artifact_chunks(artifact, chunks)

    async def _store_artifact_fulltext(self, artifact: Artifact, chunks: List[Chunk] = None) -> None:
        """Store artifact content in full-text search database.
        
        Args:
            artifact (Artifact): The artifact to store
            chunks (List[Chunk], optional): Chunks of the artifact. If None, uses the artifact's content directly.
        """
        if not self.fulltext_db:
            return
            
        try:
            documents = []

            # default use origin text
            if chunks and self.workspace_config.fulltext_db_config.config.get('use_chunk'):
                # Store each chunk as a separate document
                for chunk in chunks:
                    doc = {
                        "id": chunk.chunk_id,
                        "content": chunk.content,
                        "artifact_id": artifact.artifact_id,
                        "chunk_id": chunk.chunk_id,
                        "metadata": {
                            "artifact_type": artifact.artifact_type.value,
                            "chunk_index": chunk.chunk_metadata.chunk_index,
                            "chunk_size": chunk.chunk_metadata.chunk_size,
                            "chunk_overlap": chunk.chunk_metadata.chunk_overlap,
                            **chunk.chunk_metadata.model_dump()
                        },
                        "created_at": artifact.created_at,
                        "updated_at": artifact.updated_at
                    }
                    documents.append(doc)
            else:
                # Store the entire artifact as a single document
                content = artifact.get_embedding_text()
                if content:
                    doc = {
                        "id": artifact.artifact_id,
                        "content": content,
                        "artifact_id": artifact.artifact_id,
                        "metadata": {
                            "artifact_type": artifact.artifact_type.value,
                            "content_size": len(content),
                            **artifact.metadata
                        },
                        "created_at": artifact.created_at,
                        "updated_at": artifact.updated_at
                    }
                    documents.append(doc)
            
            if documents:
                await asyncio.to_thread(self.fulltext_db.insert, self.full_text_index, documents)
                logger.info(f"📦[FULLTEXT]✅ store_fulltext[{artifact.artifact_type}]:{artifact.artifact_id} finished, {len(documents)} documents")
            else:
                logger.warning(f"📦[FULLTEXT]⚠️ store_fulltext[{artifact.artifact_type}]:{artifact.artifact_id} no content to store")
                
        except Exception as e:
            logger.error(f"📦[FULLTEXT]❌ store_fulltext[{artifact.artifact_type}]:{artifact.artifact_id} failed: {e}")
            raise


    async def _rebuild_artifact_fulltext(self, artifact: Artifact, chunks: List[Chunk] = None) -> None:
        """Store artifact content in full-text search database.
        
        This is a public method that can be called independently to store artifacts
        in the full-text search database without rebuilding embeddings.
        
        Args:
            artifact (Artifact): The artifact to store
            chunks (List[Chunk], optional): Chunks of the artifact. If None, uses the artifact's content directly.
        """
        if not self.fulltext_db:
            logger.warning(f"📦[FULLTEXT]⚠️ fulltext_db is not enabled for artifact {artifact.artifact_id}")
            return
            
        # Delete existing full-text data for this artifact
        self.fulltext_db.delete(self.workspace_id, filter={"artifact_id": artifact.artifact_id})
        logger.info(f"📦[FULLTEXT]✅ delete_fulltext[{artifact.artifact_type}]:{artifact.artifact_id} finished")
        chunkable = artifact.get_metadata_value("chunkable")
        if chunkable and self.workspace_config.fulltext_db_config.config.get('use_chunk'):
            if not chunks:
                chunks = await self._load_artifact_chunks(artifact)
        await self._store_artifact_fulltext(artifact, chunks)

    async def delete_artifact_fulltext(self, artifact_id: str) -> None:
        """Delete artifact content from full-text search database.
        
        Args:
            artifact_id (str): The artifact ID to delete
        """
        if not self.fulltext_db:
            return
            
        try:
            self.fulltext_db.delete(self.workspace_id, filter={"artifact_id": artifact_id})
            logger.info(f"📦[FULLTEXT]✅ delete_fulltext for artifact_id: {artifact_id} finished")
        except Exception as e:
            logger.error(f"📦[FULLTEXT]❌ delete_fulltext for artifact_id: {artifact_id} failed: {e}")
            raise

    async def rebuild_fulltext(self) -> None:
        """Rebuild full-text search index for all artifacts or a specific artifact.
        
        Args:
            artifact (Artifact, optional): Specific artifact to rebuild. If None, rebuilds all artifacts.
        """
        if not self.fulltext_db:
            logger.warning("📦[FULLTEXT]⚠️ fulltext_db is not enabled")
            return

        try:
            # Rebuild for all artifacts
            logger.info("📦[FULLTEXT]🔄 rebuilding fulltext for all artifacts")
            for artifact in self.artifacts:
                await self.rebuild_artifact_fulltext(artifact)
            logger.info("📦[FULLTEXT]✅ rebuild_fulltext completed for all artifacts")

        except Exception as e:
            logger.error(f"📦[FULLTEXT]❌ rebuild_fulltext failed: {e}")
            raise

    async def rebuild_artifact_fulltext(self, artifact: Artifact):
        await self._rebuild_artifact_fulltext(artifact)
        for sub_artifact in tqdm(artifact.sublist, f"rebuild_artifact_fulltext_sublist#{artifact.artifact_id}"):
            if not sub_artifact.content:
                sub_artifact.content = self._get_file_content_by_artifact_id(artifact_id=sub_artifact.artifact_id, parent_id=artifact.artifact_id)
            await self._rebuild_artifact_fulltext(sub_artifact)

    #########################################################
    # Artifact Retrieval
    #########################################################
    @property
    def total_artifacts(self):
        results = []
        for artifact in self.artifacts:
            results += [artifact]
            if artifact.sublist:
                results += artifact.sublist
        return results

    def list_artifacts(self, artifact_ids: Optional[List[str]] = None, filter_types: Optional[List[ArtifactType]] = None, sublist=False) -> List[Artifact]:
        """
        List all artifacts in the workspace
        
        Args:
            filter_type: Optional filter type
            
        Returns:
            List of artifacts
        """
        if artifact_ids:
            return [a for a in self.artifacts if a.artifact_id in artifact_ids]
        if filter_types:
            return [a for a in self.artifacts if a.artifact_type in filter_types]
        if not sublist:
            return [a for a in self.artifacts]
        return self.total_artifacts


    
    def get_artifact(self, artifact_id: str, parent_id: str = None, load_content:bool = True, load_summary = True) -> Optional[Artifact]:
        """
        Get an artifact by its ID
        
        Args:
            artifact_id: Artifact ID
            
        Returns:
            Artifact object if found, None otherwise
        """
        if parent_id:
            parent_artifact = self.get_artifact(artifact_id=parent_id)
            if parent_artifact:
                if not parent_artifact.sublist:
                    return None
                for sub_artifact in parent_artifact.sublist:
                    if sub_artifact.artifact_id == artifact_id:
                        if load_content:
                            sub_artifact.content = self.repository.get_subaritfact_content(artifact_id, parent_id)
                        if load_summary:
                            if self.vector_db:
                                summary_result = self.vector_db.query(self.summary_vector_collection, filter={
                                    "artifact_id": sub_artifact.artifact_id
                                })
                                if summary_result and len(summary_result.docs) > 0:
                                    sub_artifact.summary = summary_result.docs[0].content
                        return sub_artifact
                return None

        return self._get_artifact(artifact_id)

    def get_next_artifact(self, artifact_id: str) -> Optional[Artifact]:
        for i, artifact in enumerate(self.artifacts):
            if artifact.artifact_id == artifact_id:
                if i + 1 >= len(self.artifacts):
                    return None
                return self.artifacts[i + 1]
            if artifact.sublist:
                for si, sub_artifact in enumerate(artifact.sublist):
                    if sub_artifact.artifact_id == artifact_id:
                        if si + 1 >= len(artifact.sublist):
                            return None
                        return artifact.sublist[si + 1]
        return None

    def get_pre_artifact(self, artifact_id: str) -> Optional[Artifact]:
        for i, artifact in enumerate(self.artifacts):
            if artifact.artifact_id == artifact_id:
                if i - 1 < 0:
                    return None
                return self.artifacts[i - 1]
            if artifact.sublist:
                for si,sub_artifact in enumerate(artifact.sublist):
                    if sub_artifact.artifact_id == artifact_id:
                        if si - 1 < 0:
                            return None
                        return artifact.sublist[si - 1]

        return None

    def _get_artifact(self, artifact_id: str) -> Optional[Artifact]:
        for artifact in self.artifacts:
            if artifact.artifact_id == artifact_id:
                return artifact
            if artifact.sublist:
                for sub_artifact in artifact.sublist:
                    if sub_artifact.artifact_id == artifact_id:
                        return sub_artifact
        return None
    
    def get_file_content_by_artifact_id(self, artifact_id: str, parent_id: str = None) -> Optional[str]:
        """
        Get concatenated content of all artifacts with the same filename.

        Args:
            artifact_id: artifact_id

        Returns:
            Raw unescaped concatenated content of all matching artifacts
        """
        artifact = self._get_artifact(artifact_id)
        if not artifact:
            return None
        return self._get_file_content_by_artifact_id(artifact.artifact_id, artifact.parent_id)

    def _get_file_content_by_artifact_id(self, artifact_id: str, parent_id: str = None) -> Optional[str]:
        if parent_id:
            return self.repository.get_subaritfact_content(artifact_id, parent_id)
        else:
            return None
    
    #########################################################
    # Hybrid Search
    #########################################################


    async def retrieve_artifacts(self, search_query: HybridSearchQuery) -> Optional[list[HybridSearchResult]]:
        """
        Retrieve an artifact by its ID

        Args:
            query: Query string
            filter_types: Optional filter types

        Returns:
            Artifact object if found, None otherwise
        """
        logger.info(f"🔍 retrieve_artifact search_query: {search_query}")

        if not self.workspace_config.hybrid_search_config.enabled:
            return []

        if not search_query:
            logger.warning("🔍 retrieve_artifact search_query is None")
            return None

        if not search_query.limit:
            search_query.limit = self.workspace_config.hybrid_search_config.top_k

        if not search_query.threshold:
            search_query.threshold = self.workspace_config.hybrid_search_config.threshold
        logger.debug(f"🔍 retrieve_artifact final search_query: {search_query}")

        candidate_results = []

        # Execute vector and fulltext search concurrently
        vector_task = asyncio.create_task(self._vector_search_artifacts(search_query))
        vector_summary_task = asyncio.create_task(self._vector_search_artifacts_by_summary(search_query))
        fulltext_task = asyncio.create_task(self._fulltext_search_artifacts(search_query))

        vector_results, fulltext_results,  vector_summary_results = await asyncio.gather(vector_task, fulltext_task, vector_summary_task)

        if vector_results:
            candidate_results += [vector_result.artifact for vector_result in vector_results]
            logger.info(f"🔍 retrieve_artifact vector_results size: {len(vector_results)}")
            for item in vector_results:
                logger.debug(f"🔍 retrieve_artifact vector_results item: {item.artifact.artifact_id}: {item.score}")

        if vector_summary_results:
            candidate_results += [vector_summary_result.artifact for vector_summary_result in vector_summary_results]
            logger.info(f"🔍 retrieve_artifact vector_results size: {len(vector_summary_results)}")
            for item in vector_summary_results:
                logger.debug(f"🔍 retrieve_artifact vector_results item: {item.artifact.artifact_id}: {item.score}")

        if fulltext_results:
            candidate_results += [fulltext_result.artifact for fulltext_result in fulltext_results]
            logger.info(f"🔍 retrieve_artifact fulltext_results size: {len(fulltext_results)}")
            for item in fulltext_results:
                logger.debug(f"🔍 retrieve_artifact fulltext_results item: {item.artifact.artifact_id}: {item.score}")

        if fulltext_results:
            candidate_results += [fulltext_result.artifact for fulltext_result in fulltext_results]
            logger.info(f"🔍 retrieve_artifact fulltext_results size: {len(fulltext_results)}")
            for item in fulltext_results:
                logger.debug(f"🔍 retrieve_artifact fulltext_results item: {item.artifact.artifact_id}: {item.score}")

        logger.info(f"🔍 retrieve_artifact candidate_results size: {len(candidate_results)}")
        rerank_results = await self._rerank_candidate_artifacts(search_query.query,candidate_results)
        for item in rerank_results:
            logger.debug(f"🔍 retrieve_artifact rerank_results item: {item.artifact.artifact_id}: {item.score}")

        # Convert rerank results to HybridSearchResults and limit to requested size
        results = [
            HybridSearchResult(artifact=result.artifact, score=result.score)
            for result in rerank_results[:search_query.limit]
        ]

        logger.info(f"🔍 retrieve_artifact results size: {len(results)}")
        return results

    async def _rerank_candidate_artifacts(self, user_message: str, candidates: list[Artifact]) -> Optional[list[RerankResult]]:
        # remove duplicate results by artifact_id
        unique_results = {}
        for result in candidates:
            unique_results[result.artifact_id] = result
            if not result.content:
                result.content = self._get_file_content_by_artifact_id(artifact_id=result.artifact_id, parent_id=result.parent_id)

        logger.info(f"🔍 _rerank_candidate_artifacts unique_results: {len(unique_results)}: {unique_results.keys()}")


        rerank_results = self.reranker.run(user_message, list(unique_results.values()), score_threshold=self.workspace_config.hybrid_search_config.threshold)
        if not rerank_results:
            logger.info(f"📊[rerank_result]: {user_message} is empty")
            return rerank_results
        for result in rerank_results:
            logger.info(f"📊[rerank_result]: {user_message} artifact: {result.artifact.artifact_id} score: {result.score}")
        return rerank_results

    async def _vector_search_artifacts(self, search_query: HybridSearchQuery) -> Optional[list[HybridSearchResult]]:
        """
        retrival artifact by chunk or origin record
        """
        query_embedding = self.embedder.embed_query(search_query.query)
        results = []
        # 2. Search vector db
        search_results = self.vector_db.search(self.default_vector_collection, [query_embedding], filter={},
                                               threshold=search_query.threshold, limit=search_query.limit)
        if not search_results:
            logger.warning("🔍 [CHUNK/ORIGIN]retrieve_artifact search_results is None")
            return None

        if not search_results.docs:
            logger.warning("🔍 [CHUNK/ORIGIN]retrieve_artifact search_results.docs is None")
            return None
        existing_artifact_ids = []

        for doc in search_results.docs:
            if not doc.metadata:
                logger.warning("🔍 [CHUNK/ORIGIN]retrieve_artifact doc.metadata is None")
                continue
            if doc.metadata.artifact_id in existing_artifact_ids:
                logger.debug(f"🔍 [CHUNK/ORIGIN]retrieve_artifact artifact_id already exists: {doc.metadata.artifact_id}")
                continue
            artifact = self.get_artifact(doc.metadata.artifact_id, parent_id=doc.metadata.parent_id)
            existing_artifact_ids.append(doc.metadata.artifact_id)
            if not artifact:
                logger.warning(f"🔍 retrieve_artifact artifact is None: {doc.metadata.artifact_id}")
                continue
            logger.debug(f"🔍 [CHUNK/ORIGIN]retrieve_artifact artifact- {artifact.artifact_id} score: {doc.score}")
            results.append(HybridSearchResult(artifact=artifact, score=doc.score))
        return results

    async def _vector_search_artifacts_by_summary(self, search_query: HybridSearchQuery) -> Optional[list[HybridSearchResult]]:
        """
        retrival artifact by summary
        """

        query_embedding = self.embedder.embed_query(search_query.query)
        results = []
        # 2. Search vector db
        search_results = self.vector_db.search(self.summary_vector_collection, [query_embedding], filter={},
                                               threshold=search_query.threshold, limit=search_query.limit)
        if not search_results:
            logger.warning("🔍 [SUMMARY] retrieve_artifact search_results is None")
            return None

        if not search_results.docs:
            logger.warning("🔍 [SUMMARY]retrieve_artifact search_results.docs is None")
            return None
        existing_artifact_ids = []

        for doc in search_results.docs:
            if not doc.metadata:
                logger.warning("🔍 [SUMMARY]retrieve_artifact doc.metadata is None")
                continue
            if doc.metadata.artifact_id in existing_artifact_ids:
                logger.debug(f"🔍 [SUMMARY]retrieve_artifact artifact_id already exists: {doc.metadata.artifact_id}")
                continue
            artifact = self.get_artifact(doc.metadata.artifact_id, parent_id=doc.metadata.parent_id)
            existing_artifact_ids.append(doc.metadata.artifact_id)
            if not artifact:
                logger.warning(f"🔍 [SUMMARY]retrieve_artifact artifact is None: {doc.metadata.artifact_id}")
                continue
            logger.debug(f"🔍 [SUMMARY]retrieve_artifact artifact- {artifact.artifact_id} score: {doc.score}")
            results.append(HybridSearchResult(artifact=artifact, score=doc.score))
        return results
    
    async def retrieve_chunk(self, search_query: ChunkSearchQuery) -> Optional[list[ChunkSearchResult]]:
        """
        Retrieve a chunk by its ID
        """
        logger.info(f"🔍 retrieve_chunk search_query: {search_query}")
        if not self.workspace_config.chunk_config.enabled:
            logger.warning("🔍 retrieve_chunk chunk_config is not enabled")
            return None
        
        results = []
        if not search_query:
            logger.warning("🔍 retrieve_chunk search_query is None")
            return None

        if not search_query.limit:
            search_query.limit = 10

        if not search_query.threshold and search_query.threshold != 0:
            search_query.threshold = 0.8

        if not search_query.pre_n and search_query.pre_n != 0:
            search_query.pre_n = 3

        if not search_query.next_n and search_query.pre_n != 0:
            search_query.next_n = 3

        chunk_query_embedding = self.embedder.embed_query(search_query.query)
        search_results = self.vector_db.search(self.default_vector_collection, [chunk_query_embedding], filter=search_query.filters, threshold=search_query.threshold, limit=search_query.limit)
        if not search_results:
            logger.warning("🔍 retrieve_chunk search_results is None")
            return None
        
        if not search_results.docs:
            logger.warning("🔍 retrieve_chunk search_results.docs is None")
            return None
        
        existing_chunk_ids = []
        for doc in search_results.docs:
            if not doc.metadata:
                logger.warning("🔍 retrieve_chunk doc.metadata is None")
                continue
            if doc.metadata.chunk_id in existing_chunk_ids:
                logger.debug(f"🔍 retrieve_chunk chunk_id already exists: {doc.metadata.chunk_id}")
                continue
            pre_n_chunks, chunk, next_n_chunks = self.repository.get_chunk_window(doc.metadata.artifact_id, parent_id = doc.metadata.parent_id, chunk_index=doc.metadata.chunk_index, pre_n=search_query.pre_n, next_n=search_query.next_n)
            if not chunk:
                logger.warning(f"🔍 retrieve_chunk chunk is None: {doc.metadata.chunk_id}")
                continue
            logger.debug(f"🔍 retrieve_chunk chunks- {pre_n_chunks} {chunk} {next_n_chunks} score: {doc.score}")
            results.append(ChunkSearchResult(pre_n_chunks=pre_n_chunks, chunk=chunk, next_n_chunks=next_n_chunks, score=doc.score))

        return results

    async def search_fulltext(self, query: str, limit: int = 10, filter_types: Optional[List[ArtifactType]] = None) -> \
    Optional[list[FulltextSearchResult]]:
        """Search artifacts using full-text search only.

        Args:
            query (str): Search query
            limit (int): Maximum number of results
            filter_types (Optional[List[ArtifactType]]): Filter by artifact types

        Returns:
            Optional[list[FulltextSearchResult]]: Search results
        """
        if not self.fulltext_db:
            logger.warning("🔍 search_fulltext fulltext_db is not enabled")
            return None

        try:
            # Build filter for artifact types if specified
            filter_dict = {}
            if filter_types:
                filter_dict["metadata.artifact_type"] = [t.value for t in filter_types]

            # Perform full-text search
            search_results = await asyncio.to_thread(
                self.fulltext_db.search,
                self.workspace_id,
                query,
                filter=filter_dict,
                limit=limit,
                offset=0
            )

            if not search_results:
                logger.info("🔍 search_fulltext no results found")
                return None

            logger.info(f"🔍 search_fulltext found {len(search_results.results)} results")
            return search_results.results

        except Exception as e:
            logger.error(f"🔍 search_fulltext failed: {e}")
            return None

    async def _fulltext_search_artifacts(self, search_query: HybridSearchQuery) -> Optional[list[HybridSearchResult]]:
        """Search artifacts using full-text search.
        
        Args:
            search_query (HybridSearchQuery): Search query
            
        Returns:
            Optional[list[HybridSearchResult]]: Search results
        """
        if not self.fulltext_db:
            return None
            
        try:
            # Build filter for artifact types if specified
            filter_dict = {}
            if search_query.filter_types:
                filter_dict["metadata.artifact_type"] = [t.value for t in search_query.filter_types]
            
            # Perform full-text search
            search_results = await asyncio.to_thread(
                self.fulltext_db.search,
                self.full_text_index,
                search_query.query,
                filter=filter_dict,
                limit=search_query.limit,
                offset=0
            )
            
            if not search_results or not search_results.results:
                logger.info("🔍 _fulltext_search_artifacts no results found")
                return None
            
            results = []
            existing_artifact_ids = []
            
            for result in search_results.results:
                if result.artifact_id in existing_artifact_ids:
                    logger.debug(f"🔍 _fulltext_search_artifacts artifact_id already exists: {result.artifact_id}")
                    continue
                    
                artifact = self.get_artifact(result.artifact_id)
                if not artifact:
                    logger.warning(f"🔍 _fulltext_search_artifacts artifact not found: {result.artifact_id}")
                    continue
                    
                logger.debug(f"🔍 _fulltext_search_artifacts artifact- {artifact.artifact_id} score: {result.score}")
                results.append(HybridSearchResult(artifact=artifact, score=result.score))
                existing_artifact_ids.append(result.artifact_id)
            
            logger.info(f"🔍 _fulltext_search_artifacts found {len(results)} results")
            return results
            
        except Exception as e:
            logger.error(f"🔍 _fulltext_search_artifacts failed: {e}")
            return None

    
    #########################################################
    # Observer Management
    #########################################################

    def add_observer(self, observer: WorkspaceObserver) -> None:
        """
        Add a workspace observer
        
        Args:
            observer: Observer object implementing WorkspaceObserver interface
        """
        if not isinstance(observer, WorkspaceObserver):
            raise TypeError("Observer must be an instance of WorkspaceObserver")
        self.observers.append(observer)

    def remove_observer(self, observer: WorkspaceObserver) -> None:
        """Remove an observer"""
        if observer in self.observers:
            self.observers.remove(observer)

    async def _notify_observers(self, operation: str, artifact: Artifact) -> List[Any]:
        """
        Notify all observers of workspace changes
        
        Args:
            operation: Type of operation (create, update, delete)
            artifact: Affected artifact
            
        Returns:
            List of results from handlers
        """
        results = []
        for observer in self.observers:
            try:
                if operation == "create":
                    result = await observer.on_create(workspace_id=self.workspace_id, artifact=artifact)
                    if result:
                        results.append(result)
                elif operation == "update":
                    result = await observer.on_update(workspace_id=self.workspace_id, artifact=artifact)
                    if result:
                        results.append(result)
                elif operation == "delete":
                    result = await observer.on_delete(workspace_id=self.workspace_id, artifact=artifact)
                    if result:
                        results.append(result)
            except Exception as e:
                print(f"Observer notification failed: {e}")
        return results


    #########################################################
    # Workspace Management
    #########################################################

    def save(self) -> None:
        """
        Save workspace state
        
        Returns:
            Workspace storage ID
        """
        workspace_data = {
            "workspace_id": self.workspace_id,
            "name": self.name,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "metadata": self.metadata,
            "artifact_ids": [a.artifact_id for a in self.artifacts],
            "artifacts": [
                {
                    "artifact_id": a.artifact_id,
                    "type": str(a.artifact_type),
                    "metadata": a.metadata,
                    # "version": a.current_version
                } for a in self.artifacts
            ]
        }

        logger.info(f"💼 save_workspace {self.workspace_id}")
        # Store workspace information with workspace_id in metadata
        self.repository.store_index(
            index_data=workspace_data
        )
    
    def _load_workspace_data(self) -> Optional[Dict[str, Any]]:
        """
        Load workspace data from index.json and artifacts/{artifact_id}/index.json

        Returns:
            Dictionary containing workspace data if exists, None otherwise
        """
        try:
            # 1. retrieve workspace index.json
            index_data = self.repository.get_index_data()
            if not index_data:
                return None
            workspace_data = index_data.get("workspace")
            if not workspace_data:
                return None

            # 2. retrieve all artifacts 
            artifacts = []
            for artifact_meta in workspace_data.get("artifacts", []):
                artifact_id = artifact_meta.get("id") or artifact_meta.get("artifact_id")
                if not artifact_id:
                    continue
                artifact_data = self.repository.retrieve_artifact(artifact_id)
                artifacts.append(Artifact.from_dict(artifact_data))

            return {
                "artifacts": artifacts,
                "metadata": workspace_data.get("metadata", {}),
                "created_at": workspace_data.get("created_at"),
                "updated_at": workspace_data.get("updated_at")
            }
        except Exception as e:
            traceback.print_exc()
            print(f"Error loading workspace data: {e}")
            return None
        
    def generate_tree_data(self) -> Dict[str, Any]:
        """
        Generate a tree structure based on artifacts and their sublist recursively.
        Returns:
            A dictionary representing the artifact tree.
        """
        def build_node(artifact, parent_id="-1", depth=1):
            node = {
                "name": artifact.metadata.get('filename', artifact.artifact_id),
                "id": artifact.artifact_id,
                "type": str(artifact.artifact_type),
                "artifactId": artifact.artifact_id,
                "parentId": parent_id,
                "depth": depth,
                "expanded": False,
                "children": []
            }
            for sub in getattr(artifact, 'sublist', []):
                node["children"].append(build_node(sub, parent_id=artifact.artifact_id, depth=depth+1))
            return node

        root = {
            "name": self.name,
            "id": "-1",
            "type": "workspace",
            "children": [build_node(artifact) for artifact in self.artifacts]
        }
        return root



from pydantic import BaseModel

from workspacex.fulltext.dbs.base import FulltextDB


class FulltextDBConfig(BaseModel):
    provider: str = "elasticsearch"
    config: dict = {}
    
    @classmethod
    def from_config(cls, config: dict):
        if not config:
            return None
        return cls(
            provider=config.get("provider", "elasticsearch"),
            config=config.get("config", {})
        )


class FulltextDBFactory:

    @staticmethod
    def get_fulltext_db(fulltext_db_config: FulltextDBConfig) -> FulltextDB:
        if fulltext_db_config.provider == "elasticsearch":
            from workspacex.fulltext.dbs.elasticsearch import ElasticsearchFulltextDB
            return ElasticsearchFulltextDB(fulltext_db_config.config)
        else:
            raise ValueError(f"Full-text database {fulltext_db_config.provider} is not supported") 


import zipfile
from pathlib import Path

from workspacex import Artifact, ArtifactType
from workspacex.artifact import AttachmentFile
from workspacex.chunk.base import ChunkConfig, ChunkerFactory
from workspacex.utils import io, pdf
from workspacex.utils.logger import logger


class ArxivArtifact(Artifact):

    def __init__(self, arxiv_id: str, page_count:int = -1, **kwargs):
        artifact_id = f"arxiv_{arxiv_id}"
        metadata = {
            "arxiv_id": arxiv_id,
            "page_count": page_count
        }
        super().__init__(artifact_id =artifact_id, content="", artifact_type=ArtifactType.ARXIV, metadata=metadata, **kwargs)

    @property
    def arxiv_id(self):
        return self.metadata.get('arxiv_id')

    @property
    def page_count(self):
        return self.metadata.get('page_count', -1)

    @page_count.setter
    def page_count(self, page_count: int):
        self.metadata['page_count'] = page_count

    @property
    def chunker(self):
        chunk_config = ChunkConfig(
            enabled=True,
            provider="markdown"
        )

        return ChunkerFactory.get_chunker(chunk_config)

    def after_chunker(self):
        self.content = ""

    async def process_arxiv(self):
        """
        Process arXiv paper: download PDF, convert to markdown, and chunk the content
        
        Returns:
            None: Updates the artifact with processed content and chunks
        """
        url = f"https://arxiv.org/pdf/{self.arxiv_id}"
        # 1. download pdf
        filename, temp_file_path = await io.download_pdf_from_url(url)
        if not filename:
            raise ValueError("❌ download_pdf_from_url failed, no filename found")
        logger.info(f"📥 Downloaded PDF from {url} to {temp_file_path}")
        self.add_attachment_file(AttachmentFile(file_name=filename, file_desc="PDF file", file_path=temp_file_path))
        logger.info(f"📎 Added PDF attachment, total attachments: {len(self.attachment_files)}")

        # 2. call api resolve pdf as markdown
        markdown_zip_filename, markdown_zip_file_path = await pdf.parse_pdf_to_zip(temp_file_path, page_count=self.page_count)
        logger.info(f"📄 Parsed PDF to markdown {markdown_zip_file_path}")
        if not markdown_zip_filename:
            raise ValueError("❌parse_pdf_to_zip No markdown zip file found")
        self.add_attachment_file(AttachmentFile(file_name=markdown_zip_filename, file_desc="Markdown zip file", file_path=markdown_zip_file_path))
        logger.info(f"📎 Added markdown zip attachment, total attachments: {len(self.attachment_files)}")

        # 3. unzip zipfile, get converted_content.markdown file
        extract_dir = Path(temp_file_path).parent / f"extracted_{self.arxiv_id}"
        extract_dir.mkdir(exist_ok=True)
        with zipfile.ZipFile(markdown_zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(extract_dir)

        # Find the markdown file in the extracted directory
        markdown_file_path = None
        for file_path in extract_dir.rglob("converted_content.markdown"):
            markdown_file_path = file_path
            break

        if not markdown_file_path:
            # Try to find .md files as fallback
            for file_path in extract_dir.rglob("*.md"):
                markdown_file_path = file_path
                break

        if not markdown_file_path:
            raise FileNotFoundError(f"❌ No markdown file found in extracted zip: {markdown_zip_file_path}")

        logger.info(f"📝 Found markdown file: {markdown_file_path}")

        # 4. Read markdown content
        with open(markdown_file_path, 'r', encoding='utf-8') as f:
            markdown_content = f.read()

        self.content = markdown_content
        logger.info(f"📖 Read markdown content, length: {len(markdown_content)} characters")

        # 5. Set the markdown content as artifact content
        # self.content = markdown_content
        self.add_attachment_file(AttachmentFile(
            file_name="result.md",
            file_desc="Markdown content file",
            file_path=str(markdown_file_path)
        ))
        logger.info(f"📎 Added markdown content attachment, total attachments: {len(self.attachment_files)}")

        self.mark_complete()

        logger.info(f"✅ Successfully processed arXiv paper {self.arxiv_id}")

    
